import logging

from galaxy import util
from galaxy.tool_shed.util.container_util import (
    generate_repository_dependencies_key_for_repository,
    get_components_from_key,
)
from galaxy.tool_shed.util.repository_util import (
    extract_components_from_tuple,
    get_tool_shed_repository_by_id,
)
from galaxy.util import UNKNOWN
from galaxy.util.tool_shed.common_util import parse_repository_dependency_tuple

log = logging.getLogger(__name__)


class Folder:
    """Container object."""

    def __init__(self, id=None, key=None, label=None, parent=None):
        self.id = id
        self.key = key
        self.label = label
        self.parent = parent
        self.current_repository_installation_errors = []
        self.current_repository_successful_installations = []
        self.description = None
        self.datatypes = []
        self.folders = []
        self.invalid_data_managers = []
        self.invalid_repository_dependencies = []
        self.invalid_tool_dependencies = []
        self.invalid_tools = []
        self.missing_test_components = []
        self.readme_files = []
        self.repository_dependencies = []
        self.repository_installation_errors = []
        self.repository_successful_installations = []
        self.test_environments = []
        self.tool_dependencies = []
        self.valid_tools = []
        self.valid_data_managers = []
        self.workflows = []

    def contains_folder(self, folder):
        for index, contained_folder in enumerate(self.folders):
            if folder == contained_folder:
                return index, contained_folder
        return 0, None

    def contains_repository_dependency(self, repository_dependency):
        listified_repository_dependency = repository_dependency.listify
        for contained_repository_dependency in self.repository_dependencies:
            if contained_repository_dependency.listify == listified_repository_dependency:
                return True
        return False

    def remove_repository_dependency(self, repository_dependency):
        listified_repository_dependency = repository_dependency.listify
        self.repository_dependencies = [
            contained_repository_dependency
            for contained_repository_dependency in self.repository_dependencies
            if contained_repository_dependency.listify != listified_repository_dependency
        ]

    def to_dict(self):
        folders = []
        if self.folders:
            for folder in self.folders:
                folders.append(folder.to_dict())
        repository_dependencies = []
        for rd in self.repository_dependencies or []:
            repository_dependencies.append(rd.to_dict())
        return {
            "description": self.description,
            "folders": folders,
            "repository_dependencies": repository_dependencies,
        }


class DataManager:
    """Data Manager object"""

    def __init__(self, id=None, name=None, version=None, data_tables=None):
        self.id = id
        self.name = name
        self.version = version
        self.data_tables = data_tables


class InvalidDataManager:
    """Invalid data Manager object"""

    def __init__(self, id=None, index=None, error=None):
        self.id = id
        self.index = index
        self.error = error


class InvalidTool:
    """Invalid tool object"""

    def __init__(
        self,
        id=None,
        tool_config=None,
        repository_id=None,
        changeset_revision=None,
        repository_installation_status=None,
    ):
        self.id = id
        self.tool_config = tool_config
        self.repository_id = repository_id
        self.changeset_revision = changeset_revision
        self.repository_installation_status = repository_installation_status


class ReadMe:
    """Readme text object"""

    def __init__(self, id=None, name=None, text=None):
        self.id = id
        self.name = name
        self.text = text


class RepositoryDependency:
    """Repository dependency object"""

    def __init__(
        self,
        id=None,
        toolshed=None,
        repository_name=None,
        repository_owner=None,
        changeset_revision=None,
        prior_installation_required=False,
        only_if_compiling_contained_td=False,
        installation_status=None,
        tool_shed_repository_id=None,
    ):
        self.id = id
        self.toolshed = toolshed
        self.repository_name = repository_name
        self.repository_owner = repository_owner
        self.changeset_revision = changeset_revision
        self.prior_installation_required = prior_installation_required
        self.only_if_compiling_contained_td = only_if_compiling_contained_td
        self.installation_status = installation_status
        self.tool_shed_repository_id = tool_shed_repository_id

    @property
    def listify(self):
        return [
            self.toolshed,
            self.repository_name,
            self.repository_owner,
            self.changeset_revision,
            self.prior_installation_required,
            self.only_if_compiling_contained_td,
        ]

    def to_dict(self):
        return {
            "toolshed": self.toolshed,
            "repository_name": self.repository_name,
            "repository_owner": self.repository_owner,
            "changeset_revision": self.changeset_revision,
        }


class Tool:
    """Tool object"""

    def __init__(
        self,
        id=None,
        tool_config=None,
        tool_id=None,
        name=None,
        description=None,
        version=None,
        profile=None,
        requirements=None,
        repository_id=None,
        changeset_revision=None,
        repository_installation_status=None,
    ):
        self.id = id
        self.tool_config = tool_config
        self.tool_id = tool_id
        self.name = name
        self.description = description
        self.version = version
        self.profile = profile
        self.requirements = requirements
        self.repository_id = repository_id
        self.changeset_revision = changeset_revision
        self.repository_installation_status = repository_installation_status


class ToolDependency:
    """Tool dependency object"""

    def __init__(
        self,
        id=None,
        name=None,
        version=None,
        type=None,
        readme=None,
        installation_status=None,
        repository_id=None,
        tool_dependency_id=None,
    ):
        self.id = id
        self.name = name
        self.version = version
        self.type = type
        self.readme = readme
        self.installation_status = installation_status
        self.repository_id = repository_id
        self.tool_dependency_id = tool_dependency_id

    @property
    def listify(self):
        return [self.name, self.version, self.type]


class UtilityContainerManager:
    def __init__(self, app):
        self.app = app

    def build_data_managers_folder(self, folder_id, data_managers, label=None):
        """Return a folder hierarchy containing Data Managers."""
        if data_managers:
            if label is None:
                label = "Data Managers"
            data_manager_id = 0
            folder_id += 1
            data_managers_root_folder = Folder(id=folder_id, key="root", label="root", parent=None)
            folder_id += 1
            key = "valid_data_managers"
            folder = Folder(id=folder_id, key=key, label=label, parent=data_managers_root_folder)
            data_managers_root_folder.folders.append(folder)
            # Insert a header row.
            data_manager_id += 1
            data_manager = DataManager(id=data_manager_id, name="Name", version="Version", data_tables="Data Tables")
            folder.valid_data_managers.append(data_manager)
            for data_manager_dict in data_managers.values():
                data_manager_id += 1
                try:
                    name = data_manager_dict.get("name", "")
                    version = data_manager_dict.get("version", "")
                    data_tables = ", ".join(data_manager_dict.get("data_tables", ""))
                except Exception as e:
                    name = str(e)
                    version = UNKNOWN
                    data_tables = UNKNOWN
                data_manager = DataManager(id=data_manager_id, name=name, version=version, data_tables=data_tables)
                folder.valid_data_managers.append(data_manager)
        else:
            data_managers_root_folder = None
        return folder_id, data_managers_root_folder

    def build_invalid_data_managers_folder(self, folder_id, data_managers, error_messages=None, label=None):
        """Return a folder hierarchy containing invalid Data Managers."""
        if data_managers or error_messages:
            if label is None:
                label = "Invalid Data Managers"
            data_manager_id = 0
            folder_id += 1
            data_managers_root_folder = Folder(id=folder_id, key="root", label="root", parent=None)
            folder_id += 1
            key = "invalid_data_managers"
            folder = Folder(id=folder_id, key=key, label=label, parent=data_managers_root_folder)
            data_managers_root_folder.folders.append(folder)
            # Insert a header row.
            data_manager_id += 1
            data_manager = InvalidDataManager(id=data_manager_id, index="Element Index", error="Error")
            folder.invalid_data_managers.append(data_manager)
            if error_messages:
                for error_message in error_messages:
                    data_manager_id += 1
                    data_manager = InvalidDataManager(id=data_manager_id, index=0, error=error_message)
                    folder.invalid_data_managers.append(data_manager)
            for data_manager_dict in data_managers:
                data_manager_id += 1
                data_manager = InvalidDataManager(
                    id=data_manager_id,
                    index=data_manager_dict.get("index", 0) + 1,
                    error=data_manager_dict.get("error_message", ""),
                )
                folder.invalid_data_managers.append(data_manager)
        else:
            data_managers_root_folder = None
        return folder_id, data_managers_root_folder

    def build_invalid_tools_folder(
        self, folder_id, invalid_tool_configs, changeset_revision, repository=None, label="Invalid tools"
    ):
        """Return a folder hierarchy containing invalid tools."""
        # TODO: Should we display invalid tools on the tool panel selection page when installing the
        # repository into Galaxy?
        if invalid_tool_configs:
            invalid_tool_id = 0
            folder_id += 1
            invalid_tools_root_folder = Folder(id=folder_id, key="root", label="root", parent=None)
            folder_id += 1
            folder = Folder(id=folder_id, key="invalid_tools", label=label, parent=invalid_tools_root_folder)
            invalid_tools_root_folder.folders.append(folder)
            for invalid_tool_config in invalid_tool_configs:
                invalid_tool_id += 1
                if repository:
                    repository_id = repository.id
                    if self.app.name == "galaxy":
                        repository_installation_status = repository.status
                    else:
                        repository_installation_status = None
                else:
                    repository_id = None
                    repository_installation_status = None
                invalid_tool = InvalidTool(
                    id=invalid_tool_id,
                    tool_config=invalid_tool_config,
                    repository_id=repository_id,
                    changeset_revision=changeset_revision,
                    repository_installation_status=repository_installation_status,
                )
                folder.invalid_tools.append(invalid_tool)
        else:
            invalid_tools_root_folder = None
        return folder_id, invalid_tools_root_folder

    def build_readme_files_folder(self, folder_id, readme_files_dict, label="Readme files"):
        """Return a folder hierarchy containing readme text files."""
        if readme_files_dict:
            readme_id = 0
            folder_id += 1
            readme_files_root_folder = Folder(id=folder_id, key="root", label="root", parent=None)
            folder_id += 1
            readme_files_folder = Folder(id=folder_id, key="readme_files", label=label, parent=readme_files_root_folder)
            multiple_readme_files = len(readme_files_dict) > 1
            readme_files_root_folder.folders.append(readme_files_folder)
            for readme_file_name, readme_file_text in readme_files_dict.items():
                readme_id += 1
                readme = ReadMe(id=readme_id, name=readme_file_name, text=readme_file_text)
                if multiple_readme_files:
                    folder_id += 1
                    folder = Folder(
                        id=folder_id, key=readme_file_name, label=readme_file_name, parent=readme_files_folder
                    )
                    folder.readme_files.append(readme)
                    readme_files_folder.folders.append(folder)
                else:
                    readme_files_folder.readme_files.append(readme)
        else:
            readme_files_root_folder = None
        return folder_id, readme_files_root_folder

    def build_repository_dependencies_folder(
        self, folder_id, repository_dependencies, label="Repository dependencies", installed=False
    ):
        """Return a folder hierarchy containing repository dependencies."""
        if repository_dependencies:
            repository_dependency_id = 0
            folder_id += 1
            # Create the root folder.
            repository_dependencies_root_folder = Folder(id=folder_id, key="root", label="root", parent=None)
            folder_id += 1
            # Create the Repository dependencies folder and add it to the root folder.
            repository_dependencies_folder_key = repository_dependencies["root_key"]
            repository_dependencies_folder = Folder(
                id=folder_id,
                key=repository_dependencies_folder_key,
                label=label,
                parent=repository_dependencies_root_folder,
            )
            del repository_dependencies["root_key"]
            # The received repository_dependencies is a dictionary with keys: 'root_key', 'description', and one or more
            # repository_dependency keys.  We want the description value associated with the repository_dependencies_folder.
            repository_dependencies_folder.description = repository_dependencies.get("description", None)
            repository_dependencies_root_folder.folders.append(repository_dependencies_folder)
            del repository_dependencies["description"]
            (
                repository_dependencies_folder,
                folder_id,
                repository_dependency_id,
            ) = self.populate_repository_dependencies_container(
                repository_dependencies_folder, repository_dependencies, folder_id, repository_dependency_id
            )
            repository_dependencies_folder = self.prune_repository_dependencies(repository_dependencies_folder)
        else:
            repository_dependencies_root_folder = None
        return folder_id, repository_dependencies_root_folder

    def build_tools_folder(
        self, folder_id, tool_dicts, repository, changeset_revision, valid=True, label="Valid tools"
    ):
        """Return a folder hierarchy containing valid tools."""
        if tool_dicts:
            container_object_tool_id = 0
            folder_id += 1
            tools_root_folder = Folder(id=folder_id, key="root", label="root", parent=None)
            folder_id += 1
            folder = Folder(id=folder_id, key="tools", label=label, parent=tools_root_folder)
            if self.app.name == "galaxy":
                folder.description = "click the name to inspect the tool metadata"
            tools_root_folder.folders.append(folder)
            # Insert a header row.
            container_object_tool_id += 1
            tool = Tool(
                id=container_object_tool_id,
                tool_config="",
                tool_id="",
                name="Name",
                description="Description",
                version="Version",
                profile="Minimum Galaxy Version",
                requirements="",
                repository_id="",
                changeset_revision="",
            )
            folder.valid_tools.append(tool)
            if repository:
                repository_id = repository.id
                if self.app.name == "galaxy":
                    repository_installation_status = repository.status
                else:
                    repository_installation_status = None
            else:
                repository_id = None
                repository_installation_status = None
            for tool_dict in tool_dicts:
                if not isinstance(tool_dict, dict):
                    # Due to some previous bug (hopefully not current) invalid tool strings may be included in the received
                    # list of tool_dicts.  For example, the picard repository metadata has 2 invalid tools in the received
                    # list of supposedly valid tools:  'rgPicardASMetrics.xml', 'rgPicardGCBiasMetrics.xml'.
                    continue
                container_object_tool_id += 1
                requirements = tool_dict.get("requirements", None)
                if requirements is not None:
                    # 'requirements': [{'version': '1.56.0', 'type': 'package', 'name': 'picard'}],
                    requirements_str = ""
                    for requirement_dict in requirements:
                        try:
                            requirement_name = str(requirement_dict.get("name", UNKNOWN))
                            requirement_type = str(requirement_dict.get("type", UNKNOWN))
                        except Exception as e:
                            requirement_name = str(e)
                            requirement_type = UNKNOWN
                        requirements_str += f"{requirement_name} ({requirement_type}), "
                    requirements_str = requirements_str.rstrip(", ")
                else:
                    requirements_str = "none"
                try:
                    tool_config = str(tool_dict.get("tool_config", "missing"))
                    tool_id = str(tool_dict.get("id", UNKNOWN))
                    name = str(tool_dict.get("name", UNKNOWN))
                    description = str(tool_dict.get("description", ""))
                    version = str(tool_dict.get("version", UNKNOWN))
                    profile = str(tool_dict.get("profile", "any"))
                except Exception as e:
                    tool_config = str(e)
                    tool_id = UNKNOWN
                    name = UNKNOWN
                    description = ""
                    version = UNKNOWN
                tool = Tool(
                    id=container_object_tool_id,
                    tool_config=tool_config,
                    tool_id=tool_id,
                    name=name,
                    description=description,
                    version=version,
                    profile=profile,
                    requirements=requirements_str,
                    repository_id=repository_id,
                    changeset_revision=changeset_revision,
                    repository_installation_status=repository_installation_status,
                )
                folder.valid_tools.append(tool)
        else:
            tools_root_folder = None
        return folder_id, tools_root_folder

    def build_tool_dependencies_folder(
        self,
        folder_id,
        tool_dependencies,
        label="Tool dependencies",
        missing=False,
        new_install=False,
        reinstalling=False,
    ):
        """Return a folder hierarchy containing tool dependencies."""
        # When we're in Galaxy (not the tool shed) and the tool dependencies are not installed or are in an error state,
        # they are considered missing.  The tool dependency status will be displayed only if a record exists for the tool
        # dependency in the Galaxy database, but the tool dependency is not installed.  The value for new_install will be
        # True only if the associated repository in being installed for the first time.  This value is used in setting the
        # container description.
        if tool_dependencies:
            tool_dependency_id = 0
            folder_id += 1
            tool_dependencies_root_folder = Folder(id=folder_id, key="root", label="root", parent=None)
            folder_id += 1
            folder = Folder(id=folder_id, key="tool_dependencies", label=label, parent=tool_dependencies_root_folder)
            if self.app.name == "galaxy":
                if new_install or reinstalling:
                    folder.description = "repository tools require handling of these dependencies"
                elif missing and not new_install and not reinstalling:
                    folder.description = "click the name to install the missing dependency"
                else:
                    folder.description = "click the name to browse the dependency installation directory"
            tool_dependencies_root_folder.folders.append(folder)
            # Insert a header row.
            tool_dependency_id += 1
            if self.app.name == "galaxy":
                tool_dependency = ToolDependency(
                    id=tool_dependency_id,
                    name="Name",
                    version="Version",
                    type="Type",
                    readme=None,
                    installation_status="Installation status",
                    repository_id=None,
                    tool_dependency_id=None,
                )
            else:
                tool_dependency = ToolDependency(
                    id=tool_dependency_id,
                    name="Name",
                    version="Version",
                    type="Type",
                    readme=None,
                    installation_status=None,
                    repository_id=None,
                    tool_dependency_id=None,
                )
            folder.tool_dependencies.append(tool_dependency)
            for dependency_key, requirements_dict in tool_dependencies.items():
                tool_dependency_id += 1
                if dependency_key in ["set_environment"]:
                    for set_environment_dict in requirements_dict:
                        try:
                            name = set_environment_dict.get("name", None)
                            type = set_environment_dict["type"]
                            repository_id = set_environment_dict.get("repository_id", None)
                            td_id = set_environment_dict.get("tool_dependency_id", None)
                        except Exception as e:
                            name = str(e)
                            type = UNKNOWN
                            repository_id = UNKNOWN
                            td_id = UNKNOWN
                        if self.app.name == "galaxy":
                            try:
                                installation_status = set_environment_dict.get("status", "Never installed")
                            except Exception as e:
                                installation_status = str(e)
                        else:
                            installation_status = None
                        tool_dependency = ToolDependency(
                            id=tool_dependency_id,
                            name=name,
                            version=None,
                            type=type,
                            readme=None,
                            installation_status=installation_status,
                            repository_id=repository_id,
                            tool_dependency_id=td_id,
                        )
                        folder.tool_dependencies.append(tool_dependency)
                else:
                    try:
                        name = requirements_dict["name"]
                        version = requirements_dict["version"]
                        type = requirements_dict["type"]
                        repository_id = requirements_dict.get("repository_id", None)
                        td_id = requirements_dict.get("tool_dependency_id", None)
                    except Exception as e:
                        name = str(e)
                        version = UNKNOWN
                        type = UNKNOWN
                        repository_id = UNKNOWN
                        td_id = UNKNOWN
                    if self.app.name == "galaxy":
                        try:
                            installation_status = requirements_dict.get("status", "Never installed")
                        except Exception as e:
                            installation_status = str(e)
                    else:
                        installation_status = None
                    tool_dependency = ToolDependency(
                        id=tool_dependency_id,
                        name=name,
                        version=version,
                        type=type,
                        readme=None,
                        installation_status=installation_status,
                        repository_id=repository_id,
                        tool_dependency_id=td_id,
                    )
                    folder.tool_dependencies.append(tool_dependency)
        else:
            tool_dependencies_root_folder = None
        return folder_id, tool_dependencies_root_folder

    def generate_repository_dependencies_folder_label_from_key(
        self,
        repository_name,
        repository_owner,
        changeset_revision,
        prior_installation_required,
        only_if_compiling_contained_td,
        key,
    ):
        """Return a repository dependency label based on the repository dependency key."""
        if self.key_is_current_repositorys_key(
            repository_name,
            repository_owner,
            changeset_revision,
            prior_installation_required,
            only_if_compiling_contained_td,
            key,
        ):
            label = "Repository dependencies"
        else:
            if util.asbool(prior_installation_required):
                prior_installation_required_str = " <i>(prior install required)</i>"
            else:
                prior_installation_required_str = ""
            label = f"Repository <b>{repository_name}</b> revision <b>{changeset_revision}</b> owned by <b>{repository_owner}</b>{prior_installation_required_str}"
        return label

    def get_components_from_repository_dependency_for_installed_repository(self, repository_dependency):
        """
        Parse a repository dependency and return components necessary for proper display
        in Galaxy on the Manage repository page.
        """
        # Default prior_installation_required and only_if_compiling_contained_td to False.
        prior_installation_required = "False"
        only_if_compiling_contained_td = "False"
        if len(repository_dependency) == 6:
            # Metadata should have been reset on this installed repository, but it wasn't.
            tool_shed_repository_id = repository_dependency[4]
            installation_status = repository_dependency[5]
            tool_shed, name, owner, changeset_revision = repository_dependency[0:4]
            repository_dependency = [
                tool_shed,
                name,
                owner,
                changeset_revision,
                prior_installation_required,
                only_if_compiling_contained_td,
            ]
        elif len(repository_dependency) == 7:
            # We have a repository dependency tuple that includes a prior_installation_required value but not a only_if_compiling_contained_td value.
            tool_shed_repository_id = repository_dependency[5]
            installation_status = repository_dependency[6]
            tool_shed, name, owner, changeset_revision, prior_installation_required = repository_dependency[0:5]
            repository_dependency = [
                tool_shed,
                name,
                owner,
                changeset_revision,
                prior_installation_required,
                only_if_compiling_contained_td,
            ]
        elif len(repository_dependency) == 8:
            # We have a repository dependency tuple that includes both a prior_installation_required value
            # and a only_if_compiling_contained_td value.
            tool_shed_repository_id = repository_dependency[6]
            installation_status = repository_dependency[7]
            repository_dependency = repository_dependency[0:6]
        else:
            tool_shed_repository_id = None
            installation_status = UNKNOWN
        if tool_shed_repository_id:
            tool_shed_repository = get_tool_shed_repository_by_id(
                self.app, self.app.security.encode_id(tool_shed_repository_id)
            )
            if tool_shed_repository:
                if tool_shed_repository.missing_repository_dependencies:
                    installation_status = f"{installation_status}, missing repository dependencies"
                elif tool_shed_repository.missing_tool_dependencies:
                    installation_status = f"{installation_status}, missing tool dependencies"
        return tool_shed_repository_id, installation_status, repository_dependency

    def get_folder(self, folder, key):
        if folder.key == key:
            return folder
        for sub_folder in folder.folders:
            return self.get_folder(sub_folder, key)
        return None

    def handle_repository_dependencies_container_entry(
        self, repository_dependencies_folder, rd_key, rd_value, folder_id, repository_dependency_id, folder_keys
    ):
        repository_components_tuple = get_components_from_key(rd_key)
        components_list = extract_components_from_tuple(repository_components_tuple)
        toolshed, repository_name, repository_owner, changeset_revision = components_list[0:4]
        # For backward compatibility to the 12/20/12 Galaxy release.
        if len(components_list) == 4:
            prior_installation_required = "False"
            only_if_compiling_contained_td = "False"
        elif len(components_list) == 5:
            prior_installation_required = components_list[4]
            only_if_compiling_contained_td = "False"
        elif len(components_list) == 6:
            prior_installation_required = components_list[4]
            only_if_compiling_contained_td = components_list[5]
        folder = self.get_folder(repository_dependencies_folder, rd_key)
        label = self.generate_repository_dependencies_folder_label_from_key(
            repository_name,
            repository_owner,
            changeset_revision,
            prior_installation_required,
            only_if_compiling_contained_td,
            repository_dependencies_folder.key,
        )
        if folder:
            if rd_key not in folder_keys:
                folder_id += 1
                sub_folder = Folder(id=folder_id, key=rd_key, label=label, parent=folder)
                folder.folders.append(sub_folder)
            else:
                sub_folder = folder
        else:
            folder_id += 1
            sub_folder = Folder(id=folder_id, key=rd_key, label=label, parent=repository_dependencies_folder)
            repository_dependencies_folder.folders.append(sub_folder)
        if self.app.name == "galaxy":
            # Insert a header row.
            repository_dependency_id += 1
            repository_dependency = RepositoryDependency(
                id=repository_dependency_id,
                repository_name="Name",
                changeset_revision="Revision",
                repository_owner="Owner",
                installation_status="Installation status",
            )
            # Insert the header row into the folder.
            sub_folder.repository_dependencies.append(repository_dependency)
        for repository_dependency in rd_value:
            if self.app.name == "galaxy":
                (
                    tool_shed_repository_id,
                    installation_status,
                    repository_dependency,
                ) = self.get_components_from_repository_dependency_for_installed_repository(repository_dependency)
            else:
                tool_shed_repository_id = None
                installation_status = None
            can_create_dependency = not self.is_subfolder_of(sub_folder, repository_dependency)
            if can_create_dependency:
                (
                    toolshed,
                    repository_name,
                    repository_owner,
                    changeset_revision,
                    prior_installation_required,
                    only_if_compiling_contained_td,
                ) = parse_repository_dependency_tuple(repository_dependency)
                repository_dependency_id += 1
                repository_dependency = RepositoryDependency(
                    id=repository_dependency_id,
                    toolshed=toolshed,
                    repository_name=repository_name,
                    repository_owner=repository_owner,
                    changeset_revision=changeset_revision,
                    prior_installation_required=util.asbool(prior_installation_required),
                    only_if_compiling_contained_td=util.asbool(only_if_compiling_contained_td),
                    installation_status=installation_status,
                    tool_shed_repository_id=tool_shed_repository_id,
                )
                # Insert the repository_dependency into the folder.
                sub_folder.repository_dependencies.append(repository_dependency)
        return repository_dependencies_folder, folder_id, repository_dependency_id

    def is_subfolder_of(self, folder, repository_dependency):
        (
            toolshed,
            repository_name,
            repository_owner,
            changeset_revision,
            prior_installation_required,
            only_if_compiling_contained_td,
        ) = parse_repository_dependency_tuple(repository_dependency)
        key = generate_repository_dependencies_key_for_repository(
            toolshed,
            repository_name,
            repository_owner,
            changeset_revision,
            prior_installation_required,
            only_if_compiling_contained_td,
        )
        for sub_folder in folder.folders:
            if key == sub_folder.key:
                return True
        return False

    def key_is_current_repositorys_key(
        self,
        repository_name,
        repository_owner,
        changeset_revision,
        prior_installation_required,
        only_if_compiling_contained_td,
        key,
    ):
        repository_components_tuple = get_components_from_key(key)
        components_list = extract_components_from_tuple(repository_components_tuple)
        toolshed, key_name, key_owner, key_changeset_revision = components_list[0:4]
        # For backward compatibility to the 12/20/12 Galaxy release.
        if len(components_list) == 4:
            key_prior_installation_required = "False"
            key_only_if_compiling_contained_td = "False"
        elif len(components_list) == 5:
            key_prior_installation_required = components_list[4]
            key_only_if_compiling_contained_td = "False"
        elif len(components_list) == 6:
            key_prior_installation_required = components_list[4]
            key_only_if_compiling_contained_td = components_list[5]
        if (
            repository_name == key_name
            and repository_owner == key_owner
            and changeset_revision == key_changeset_revision
            and prior_installation_required == key_prior_installation_required
            and only_if_compiling_contained_td == key_only_if_compiling_contained_td
        ):
            return True
        return False

    def populate_repository_dependencies_container(
        self, repository_dependencies_folder, repository_dependencies, folder_id, repository_dependency_id
    ):
        folder_keys = []
        for key in repository_dependencies.keys():
            if key not in folder_keys:
                folder_keys.append(key)
        for key, value in repository_dependencies.items():
            (
                repository_dependencies_folder,
                folder_id,
                repository_dependency_id,
            ) = self.handle_repository_dependencies_container_entry(
                repository_dependencies_folder, key, value, folder_id, repository_dependency_id, folder_keys
            )
        return repository_dependencies_folder, folder_id, repository_dependency_id

    def prune_folder(self, folder, repository_dependency):
        listified_repository_dependency = repository_dependency.listify
        if self.is_subfolder_of(folder, listified_repository_dependency):
            folder.repository_dependencies.remove(repository_dependency)

    def prune_repository_dependencies(self, folder):
        """
        Since the object used to generate a repository dependencies container is a dictionary
        and not an OrderedDict() (it must be json-serialize-able), the order in which the dictionary
        is processed to create the container sometimes results in repository dependency entries
        in a folder that also includes the repository dependency as a sub-folder (if the repository
        dependency has its own repository dependency).  This method will remove all repository
        dependencies from folder that are also sub-folders of folder.
        """
        repository_dependencies = [rd for rd in folder.repository_dependencies]
        for repository_dependency in repository_dependencies:
            self.prune_folder(folder, repository_dependency)
        for sub_folder in folder.folders:
            return self.prune_repository_dependencies(sub_folder)
        return folder



import logging
import os
import time
import traceback
from datetime import timedelta

try:
    import pbs

    PBS_IMPORT_MESSAGE = None
except ImportError as exc:
    pbs = None
    PBS_IMPORT_MESSAGE = (
        "The Python pbs-python package is required to use "
        "this feature, please install it or correct the "
        "following error:\nImportError %s" % str(exc)
    )

from galaxy import (
    model,
    util,
)
from galaxy.jobs import JobDestination
from galaxy.jobs.runners import (
    AsynchronousJobRunner,
    AsynchronousJobState,
)
from galaxy.util.bunch import Bunch

log = logging.getLogger(__name__)

__all__ = ("PBSJobRunner",)

CLUSTER_ERROR_MESSAGE = "Job cannot be completed due to a cluster error, please retry it later: %s"

# The last two lines execute the command and then retrieve the command's
# exit code ($?) and write it to a file.
pbs_symlink_template = """
for dataset in %s; do
    dir=`dirname $dataset`
    file=`basename $dataset`
    [ ! -d $dir ] && mkdir -p $dir
    [ ! -e $dataset ] && ln -s %s/$file $dataset
done
mkdir -p %s
"""

PBS_ARGMAP = {
    "destination": "-q",
    "Execution_Time": "-a",
    "Account_Name": "-A",
    "Checkpoint": "-c",
    "Error_Path": "-e",
    "Group_List": "-g",
    "Hold_Types": "-h",
    "Join_Paths": "-j",
    "Keep_Files": "-k",
    "Resource_List": "-l",
    "Mail_Points": "-m",
    "Mail_Users": "-M",
    "Job_Name": "-N",
    "Output_Path": "-o",
    "Priority": "-p",
    "Rerunable": "-r",
    "Shell_Path_List": "-S",
    "job_array_request": "-t",
    "User_List": "-u",
    "Variable_List": "-v",
}

# From pbs' pbs_job.h
JOB_EXIT_STATUS = {
    0: "job exec successful",
    -1: "job exec failed, before files, no retry",
    -2: "job exec failed, after files, no retry",
    -3: "job execution failed, do retry",
    -4: "job aborted on MOM initialization",
    -5: "job aborted on MOM init, chkpt, no migrate",
    -6: "job aborted on MOM init, chkpt, ok migrate",
    -7: "job restart failed",
    -8: "exec() of user command failed",
    -9: "could not create/open stdout stderr files",
    -10: "job exceeded a memory limit",
    -11: "job exceeded a walltime limit",
    -12: "job exceeded a cpu time limit",
}


class PBSJobRunner(AsynchronousJobRunner):
    """
    Job runner backed by a finite pool of worker threads. FIFO scheduling
    """

    runner_name = "PBSRunner"

    def __init__(self, app, nworkers):
        """Start the job runner"""
        # Check if PBS was importable, fail if not
        assert pbs is not None, PBS_IMPORT_MESSAGE
        if app.config.pbs_application_server and app.config.outputs_to_working_directory:
            raise Exception(
                "pbs_application_server (file staging) and outputs_to_working_directory options are mutually exclusive"
            )

        # Set the default server during startup
        self.__default_pbs_server = None
        self.default_pbs_server  # noqa: B018 this is a method with a property decorator, so this causes the default server to be set

        # Proceed with general initialization
        super().__init__(app, nworkers)

    @property
    def default_pbs_server(self):
        if self.__default_pbs_server is None:
            self.__default_pbs_server = pbs.pbs_default()
            log.debug(f"Set default PBS server to {self.default_pbs_server}")
        return self.__default_pbs_server

    def url_to_destination(self, url):
        """Convert a legacy URL to a job destination"""

        if not url:
            return

        # Determine the PBS server
        url_split = url.split("/")
        server = url_split[2]
        if server == "":
            server = self.default_pbs_server
        if server is None:
            raise Exception("Could not find TORQUE server")

        # Determine the queue, set the PBS destination (not the same thing as a Galaxy job destination)
        pbs_destination = f"@{server}"
        if (pbs_queue := url_split[3] or None) is not None:
            pbs_destination = f"{pbs_queue}{pbs_destination}"

        params = dict(destination=pbs_destination)

        # Determine the args (long-format args were never supported in URLs so they are not supported here)
        try:
            opts = url.split("/")[4].strip().lstrip("-").split(" -")
            assert opts != [""]
            # stripping the - comes later (in parse_destination_params)
            for i, opt in enumerate(opts):
                opts[i] = f"-{opt}"
        except Exception:
            opts = []
        for opt in opts:
            param, value = opt.split(None, 1)
            params[param] = value

        log.debug(f"Converted URL '{url}' to destination runner=pbs, params={params}")

        # Create a dynamic JobDestination
        return JobDestination(runner="pbs", params=params)

    def parse_destination_params(self, params):
        """A wrapper method around __args_to_attrs() that allow administrators to define PBS
        params as either command-line options (as in ``qsub(1B)``) or more human-readable "long"
        args (as in ``pbs_submit(3B)``).

        :returns: list of dicts -- The dicts map directly to pbs attropl structs (see ``pbs_submit(3B)``)
        """
        args = {}
        for arg, value in params.items():
            try:
                if not arg.startswith("-"):
                    arg = PBS_ARGMAP[arg]
                arg = arg.lstrip("-")
                args[arg] = value
            except Exception:
                log.warning(f"Unrecognized long argument in destination params: {arg}")
        return self.__args_to_attrs(args)

    # Internal stuff
    def __args_to_attrs(self, args):
        """Convert a list of PBS command-line args (as in ``qsub(1B)``) to PBS' internal attribute representations.

        :returns: list of dicts -- The dicts map directly to pbs attropl structs (see ``pbs_submit(3B)``)
        """
        rval = []
        for arg, value in args.items():
            if arg == "l":
                resource_attrs = value.split(",")
                for res, val in [a.split("=", 1) for a in resource_attrs]:
                    rval.append(dict(name=pbs.ATTR_l, value=val, resource=res))
            else:
                try:
                    rval.append(dict(name=getattr(pbs, f"ATTR_{arg}"), value=value))
                except AttributeError as e:
                    raise Exception(f"Invalid parameter '{arg}': {e}")
        return rval

    def __get_pbs_server(self, job_destination_params):
        if job_destination_params is None:
            return None
        return job_destination_params["destination"].split("@")[-1]

    def queue_job(self, job_wrapper):
        """Create PBS script for a job and submit it to the PBS queue"""
        # prepare the job
        if not self.prepare_job(job_wrapper, include_metadata=not (self.app.config.pbs_stage_path)):
            return

        job_destination = job_wrapper.job_destination

        # Determine the job's PBS destination (server/queue) and options from the job destination definition
        pbs_queue_name = None
        pbs_server_name = self.default_pbs_server
        pbs_options = []
        if "-q" in job_destination.params and "destination" not in job_destination.params:
            job_destination.params["destination"] = job_destination.params.pop("-q")
        if "destination" in job_destination.params:
            if "@" in job_destination.params["destination"]:
                # Destination includes a server
                pbs_queue_name, pbs_server_name = job_destination.params["destination"].split("@")
                if pbs_queue_name == "":
                    # e.g. `qsub -q @server`
                    pbs_queue_name = None
            else:
                # Destination is just a queue
                pbs_queue_name = job_destination.params["destination"]
            job_destination.params.pop("destination")

        # Parse PBS params
        pbs_options = self.parse_destination_params(job_destination.params)

        # Explicitly set the determined PBS destination in the persisted job destination for recovery
        job_destination.params["destination"] = f"{pbs_queue_name or ''}@{pbs_server_name}"

        c = pbs.pbs_connect(util.smart_str(pbs_server_name))
        if c <= 0:
            errno, text = pbs.error()
            job_wrapper.fail("Unable to queue job for execution.  Resubmitting the job may succeed.")
            log.error(f"Connection to PBS server for submit failed: {errno}: {text}")
            return

        # define job attributes
        ofile = f"{job_wrapper.working_directory}/{job_wrapper.job_id}.o"
        efile = f"{job_wrapper.working_directory}/{job_wrapper.job_id}.e"
        ecfile = f"{job_wrapper.working_directory}/{job_wrapper.job_id}.ec"

        output_fnames = job_wrapper.job_io.get_output_fnames()

        # If an application server is set, we're staging
        if self.app.config.pbs_application_server:
            pbs_ofile = f"{self.app.config.pbs_application_server}:{ofile}"
            pbs_efile = f"{self.app.config.pbs_application_server}:{efile}"
            output_files = [str(o) for o in output_fnames]
            output_files.append(ecfile)
            stagein = self.get_stage_in_out(job_wrapper.job_io.get_input_fnames() + output_files, symlink=True)
            stageout = self.get_stage_in_out(output_files)
            attrs = [
                dict(name=pbs.ATTR_o, value=pbs_ofile),
                dict(name=pbs.ATTR_e, value=pbs_efile),
                dict(name=pbs.ATTR_stagein, value=stagein),
                dict(name=pbs.ATTR_stageout, value=stageout),
            ]
        # If not, we're using NFS
        else:
            attrs = [
                dict(name=pbs.ATTR_o, value=ofile),
                dict(name=pbs.ATTR_e, value=efile),
            ]

        # define PBS job options
        attrs.append(dict(name=pbs.ATTR_N, value=str(f"{job_wrapper.job_id}_{job_wrapper.tool.id}_{job_wrapper.user}")))
        job_attrs = pbs.new_attropl(len(attrs) + len(pbs_options))
        for i, attr in enumerate(attrs + pbs_options):
            job_attrs[i].name = attr["name"]
            job_attrs[i].value = attr["value"]
            if "resource" in attr:
                job_attrs[i].resource = attr["resource"]
        exec_dir = os.path.abspath(job_wrapper.working_directory)

        # write the job script
        if self.app.config.pbs_stage_path != "":
            # touch the ecfile so that it gets staged
            with open(ecfile, "a"):
                os.utime(ecfile, None)

            stage_commands = pbs_symlink_template % (
                " ".join(job_wrapper.job_io.get_input_fnames() + output_files),
                self.app.config.pbs_stage_path,
                exec_dir,
            )
        else:
            stage_commands = ""

        env_setup_commands = [stage_commands]
        script = self.get_job_file(
            job_wrapper, exit_code_path=ecfile, env_setup_commands=env_setup_commands, shell=job_wrapper.shell
        )
        job_file = f"{job_wrapper.working_directory}/{job_wrapper.job_id}.sh"
        self.write_executable_script(job_file, script, job_io=job_wrapper.job_io)
        # job was deleted while we were preparing it
        if job_wrapper.get_state() in (model.Job.states.DELETED, model.Job.states.STOPPED):
            log.debug(f"Job {job_wrapper.job_id} deleted/stopped by user before it entered the PBS queue")
            pbs.pbs_disconnect(c)
            if job_wrapper.cleanup_job in ("always", "onsuccess"):
                self.cleanup((ofile, efile, ecfile, job_file))
                job_wrapper.cleanup()
            return

        # submit
        # The job tag includes the job and the task identifier
        # (if a TaskWrapper was passed in):
        galaxy_job_id = job_wrapper.get_id_tag()
        log.debug(f"({galaxy_job_id}) submitting file {job_file}")

        tries = 0
        while tries < 5:
            job_id = pbs.pbs_submit(c, job_attrs, job_file, pbs_queue_name, None)
            tries += 1
            if job_id:
                pbs.pbs_disconnect(c)
                break
            errno, text = pbs.error()
            log.warning("(%s) pbs_submit failed (try %d/5), PBS error %d: %s" % (galaxy_job_id, tries, errno, text))
            time.sleep(2)
        else:
            log.error(f"({galaxy_job_id}) All attempts to submit job failed")
            job_wrapper.fail("Unable to run this job due to a cluster error, please retry it later")
            return

        if pbs_queue_name is None:
            log.debug(f"({galaxy_job_id}) queued in default queue as {job_id}")
        else:
            log.debug(f"({galaxy_job_id}) queued in {pbs_queue_name} queue as {job_id}")

        # persist destination
        job_wrapper.set_job_destination(job_destination, job_id)

        # Store PBS related state information for job
        job_state = AsynchronousJobState(
            job_wrapper=job_wrapper,
            job_id=job_id,
            exit_code_file=ecfile,
            job_destination=job_destination,
            job_file=job_file,
            output_file=ofile,
            error_file=efile,
        )
        job_state.old_state = "N"
        job_state.running = False

        # Add to our 'queue' of jobs to monitor
        self.monitor_queue.put(job_state)

    def check_watched_items(self):
        """
        Called by the monitor thread to look at each watched job and deal
        with state changes.
        """
        new_watched = []
        # reduce pbs load by batching status queries
        (failures, statuses) = self.check_all_jobs()
        for pbs_job_state in self.watched:
            job_id = pbs_job_state.job_id
            galaxy_job_id = pbs_job_state.job_wrapper.get_id_tag()
            old_state = pbs_job_state.old_state
            pbs_server_name = self.__get_pbs_server(pbs_job_state.job_destination.params)
            if pbs_server_name in failures:
                log.debug(f"({galaxy_job_id}/{job_id}) Skipping state check because PBS server connection failed")
                new_watched.append(pbs_job_state)
                continue
            try:
                status = statuses[job_id]
            except KeyError:
                if pbs_job_state.job_wrapper.get_state() == model.Job.states.DELETED:
                    continue
                try:
                    # Recheck to make sure it wasn't a communication problem
                    self.check_single_job(pbs_server_name, job_id)
                    log.warning(
                        f"({galaxy_job_id}/{job_id}) PBS job was not in state check list, but was found with individual state check"
                    )
                    new_watched.append(pbs_job_state)
                except Exception:
                    errno, text = pbs.error()
                    if errno == 15001:
                        # 15001 == job not in queue
                        log.debug(f"({galaxy_job_id}/{job_id}) PBS job has left queue")
                        self.work_queue.put((self.finish_job, pbs_job_state))
                    else:
                        # Unhandled error, continue to monitor
                        log.info(
                            "(%s/%s) PBS state check resulted in error (%d): %s" % (galaxy_job_id, job_id, errno, text)
                        )
                        new_watched.append(pbs_job_state)
                continue
            if status.job_state != old_state:
                log.debug(f"({galaxy_job_id}/{job_id}) PBS job state changed from {old_state} to {status.job_state}")
            if status.job_state == "R" and not pbs_job_state.running:
                pbs_job_state.running = True
                pbs_job_state.job_wrapper.change_state(model.Job.states.RUNNING)
            if status.job_state == "R" and status.get("resources_used", False):
                # resources_used may not be in the status for new jobs
                h, m, s = (int(i) for i in status.resources_used.walltime.split(":"))
                runtime = timedelta(0, s, 0, 0, m, h)
                if pbs_job_state.check_limits(runtime=runtime):
                    self.work_queue.put((self.fail_job, pbs_job_state))
                    continue
            elif status.job_state == "C":
                # "keep_completed" is enabled in PBS, so try to check exit status
                try:
                    assert (
                        int(status.exit_status) == 0
                        or pbs_job_state.job_wrapper.get_state() == model.Job.states.STOPPED
                    )
                    log.debug(f"({galaxy_job_id}/{job_id}) PBS job has completed successfully")
                except AssertionError:
                    exit_status = int(status.exit_status)
                    error_message = JOB_EXIT_STATUS.get(exit_status, f"Unknown error: {status.exit_status}")
                    pbs_job_state.fail_message = CLUSTER_ERROR_MESSAGE % error_message
                    log.error(f"({galaxy_job_id}/{job_id}) PBS job failed: {error_message}")
                    pbs_job_state.stop_job = False
                    self.work_queue.put((self.fail_job, pbs_job_state))
                    continue
                except AttributeError:
                    # No exit_status, can't verify proper completion so we just have to assume success.
                    log.debug(f"({galaxy_job_id}/{job_id}) PBS job has completed")
                self.work_queue.put((self.finish_job, pbs_job_state))
                continue
            pbs_job_state.old_state = status.job_state
            new_watched.append(pbs_job_state)
        # Replace the watch list with the updated version
        self.watched = new_watched

    def check_all_jobs(self):
        """
        Returns a list of servers that failed to be contacted and a dict
        of "job_id : status" pairs (where status is a bunchified version
        of the API's structure.
        """
        servers = []
        failures = []
        statuses = {}
        for pbs_job_state in self.watched:
            pbs_server_name = self.__get_pbs_server(pbs_job_state.job_destination.params)
            if pbs_server_name not in servers:
                servers.append(pbs_server_name)
            pbs_job_state.check_count += 1
        for pbs_server_name in servers:
            c = pbs.pbs_connect(util.smart_str(pbs_server_name))
            if c <= 0:
                log.debug(f"connection to PBS server {pbs_server_name} for state check failed")
                failures.append(pbs_server_name)
                continue
            stat_attrl = pbs.new_attrl(3)
            stat_attrl[0].name = pbs.ATTR_state
            stat_attrl[1].name = pbs.ATTR_used
            stat_attrl[2].name = pbs.ATTR_exitstat
            jobs = pbs.pbs_statjob(c, None, stat_attrl, None)
            pbs.pbs_disconnect(c)
            statuses.update(self.convert_statjob_to_bunches(jobs))
        return (failures, statuses)

    def convert_statjob_to_bunches(self, statjob_out):
        statuses = {}
        for job in statjob_out:
            status = {}
            for attrib in job.attribs:
                if attrib.resource is None:
                    status[attrib.name] = attrib.value
                else:
                    if attrib.name not in status:
                        status[attrib.name] = Bunch()
                    status[attrib.name][attrib.resource] = attrib.value
            statuses[job.name] = Bunch(**status)
        return statuses

    def check_single_job(self, pbs_server_name, job_id):
        """
        Returns the state of a single job, used to make sure a job is
        really dead.
        """
        c = pbs.pbs_connect(util.smart_str(pbs_server_name))
        if c <= 0:
            log.debug(f"connection to PBS server {pbs_server_name} for state check failed")
            return None
        stat_attrl = pbs.new_attrl(1)
        stat_attrl[0].name = pbs.ATTR_state
        jobs = pbs.pbs_statjob(c, job_id, stat_attrl, None)
        pbs.pbs_disconnect(c)
        return jobs[0].attribs[0].value

    def get_stage_in_out(self, fnames, symlink=False):
        """Convenience function to create a stagein/stageout list"""
        stage = ""
        for fname in fnames:
            if os.access(fname, os.R_OK):
                if stage:
                    stage += ","
                # pathnames are now absolute
                if symlink and self.app.config.pbs_stage_path:
                    stage_name = os.path.join(self.app.config.pbs_stage_path, os.path.split(fname)[1])
                else:
                    stage_name = fname
                stage += f"{stage_name}@{self.app.config.pbs_dataset_server}:{fname}"
        return stage

    def stop_job(self, job_wrapper):
        """Attempts to delete a job from the PBS queue"""
        job = job_wrapper.get_job()
        job_id = job.get_job_runner_external_id().encode("utf-8")
        job_tag = f"({job.get_id_tag()}/{job_id})"
        log.debug(f"{job_tag} Stopping PBS job")

        # Declare the connection handle c so that it can be cleaned up:
        c = None

        try:
            pbs_server_name = self.__get_pbs_server(job.destination_params)
            if pbs_server_name is None:
                log.debug("(%s) Job queued but no destination stored in job params, cannot delete" % job_tag)
                return
            c = pbs.pbs_connect(util.smart_str(pbs_server_name))
            if c <= 0:
                log.debug(f"({job_tag}) Connection to PBS server for job delete failed")
                return
            pbs.pbs_deljob(c, job_id, "")
            log.debug(f"{job_tag} Removed from PBS queue before job completion")
        except Exception:
            e = traceback.format_exc()
            log.debug(f"{job_tag} Unable to stop job: {e}")
        finally:
            # Cleanup: disconnect from the server.
            if None is not c:
                pbs.pbs_disconnect(c)

    def recover(self, job, job_wrapper):
        """Recovers jobs stuck in the queued/running state when Galaxy started"""
        job_id = job.get_job_runner_external_id()
        pbs_job_state = AsynchronousJobState(
            job_wrapper=job_wrapper,
            job_id=job_id,
            job_file=f"{job_wrapper.working_directory}/{job.id}.sh",
            output_file=f"{job_wrapper.working_directory}/{job.id}.o",
            error_file=f"{job_wrapper.working_directory}/{job.id}.e",
            exit_code_file=f"{job_wrapper.working_directory}/{job.id}.ec",
            job_destination=job_wrapper.job_destination,
        )
        pbs_job_state.runner_url = job_wrapper.get_job_runner_url()
        job_wrapper.command_line = job.command_line
        if job.state in (model.Job.states.RUNNING, model.Job.states.STOPPED):
            log.debug(
                f"({job.id}/{job.get_job_runner_external_id()}) is still in {job.state} state, adding to the PBS queue"
            )
            pbs_job_state.old_state = "R"
            pbs_job_state.running = True
            self.monitor_queue.put(pbs_job_state)
        elif job.state == model.Job.states.QUEUED:
            log.debug(
                f"({job.id}/{job.get_job_runner_external_id()}) is still in PBS queued state, adding to the PBS queue"
            )
            pbs_job_state.old_state = "Q"
            pbs_job_state.running = False
            self.monitor_queue.put(pbs_job_state)



import copy
import itertools
import logging
from collections import namedtuple
from typing import (
    Any,
    Dict,
    List,
    Optional,
    Tuple,
)

from galaxy import (
    exceptions,
    util,
)
from galaxy.model import HistoryDatasetCollectionAssociation
from galaxy.model.dataset_collections import (
    matching,
    subcollections,
)
from galaxy.util import permutations
from . import visit_input_values

log = logging.getLogger(__name__)

WorkflowParameterExpansion = namedtuple(
    "WorkflowParameterExpansion", ["param_combinations", "param_keys", "input_combinations"]
)


class ParamKey:
    def __init__(self, step_id, key):
        self.step_id = step_id
        self.key = key


class InputKey:
    def __init__(self, input_id):
        self.input_id = input_id


def expand_workflow_inputs(param_inputs, inputs=None):
    """
    Expands incoming encoded multiple payloads, into the set of all individual payload combinations
    >>> expansion = expand_workflow_inputs({'1': {'input': {'batch': True, 'product': True, 'values': [{'hid': '1'}, {'hid': '2'}] }}})
    >>> print(["%s" % (p['1']['input']['hid']) for p in expansion.param_combinations])
    ['1', '2']
    >>> expansion = expand_workflow_inputs({'1': {'input': {'batch': True, 'values': [{'hid': '1'}, {'hid': '2'}] }}})
    >>> print(["%s" % (p['1']['input']['hid']) for p in expansion.param_combinations])
    ['1', '2']
    >>> expansion = expand_workflow_inputs({'1': {'input': {'batch': True, 'values': [{'hid': '1'}, {'hid': '2'}] }}, '2': {'input': {'batch': True, 'values': [{'hid': '3'}, {'hid': '4'}] }}})
    >>> print(["%s%s" % (p['1']['input']['hid'], p['2']['input']['hid']) for p in expansion.param_combinations])
    ['13', '24']
    >>> expansion = expand_workflow_inputs({'1': {'input': {'batch': True, 'product': True, 'values': [{'hid': '1'}, {'hid': '2'}] }}, '2': {'input': {'batch': True, 'values': [{'hid': '3'}, {'hid': '4'}, {'hid': '5'}] }}})
    >>> print(["%s%s" % (p['1']['input']['hid'], p['2']['input']['hid']) for p in expansion.param_combinations])
    ['13', '23', '14', '24', '15', '25']
    >>> expansion = expand_workflow_inputs({'1': {'input': {'batch': True, 'product': True, 'values': [{'hid': '1'}, {'hid': '2'}] }}, '2': {'input': {'batch': True, 'product': True, 'values': [{'hid': '3'}, {'hid': '4'}, {'hid': '5'}] }}, '3': {'input': {'batch': True, 'product': True, 'values': [{'hid': '6'}, {'hid': '7'}, {'hid': '8'}] }}})
    >>> print(["%s%s%s" % (p['1']['input']['hid'], p['2']['input']['hid'], p['3']['input']['hid']) for p in expansion.param_combinations])
    ['136', '137', '138', '146', '147', '148', '156', '157', '158', '236', '237', '238', '246', '247', '248', '256', '257', '258']
    >>> expansion = expand_workflow_inputs(None, inputs={'myinput': {'batch': True, 'product': True, 'values': [{'hid': '1'}, {'hid': '2'}] }})
    >>> print(["%s" % (p['myinput']['hid']) for p in expansion.input_combinations])
    ['1', '2']
    """
    param_inputs = param_inputs or {}
    inputs = inputs or {}

    linked_n = None
    linked = []
    product = []
    linked_keys = []
    product_keys = []

    def is_batch(value):
        return (
            isinstance(value, dict)
            and "batch" in value
            and value["batch"] is True
            and "values" in value
            and isinstance(value["values"], list)
        )

    for step_id, step in sorted(param_inputs.items()):
        for key, value in sorted(step.items()):
            if is_batch(value):
                nval = len(value["values"])
                if "product" in value and value["product"] is True:
                    product.append(value["values"])
                    product_keys.append(ParamKey(step_id, key))
                else:
                    if linked_n is None:
                        linked_n = nval
                    elif linked_n != nval or nval == 0:
                        raise exceptions.RequestParameterInvalidException(
                            "Failed to match linked batch selections. Please select equal number of data files."
                        )
                    linked.append(value["values"])
                    linked_keys.append(ParamKey(step_id, key))

    # Force it to a list to allow modification...
    input_items = list(inputs.items())
    for input_id, value in input_items:
        if is_batch(value):
            nval = len(value["values"])
            if "product" in value and value["product"] is True:
                product.append(value["values"])
                product_keys.append(InputKey(input_id))
            else:
                if linked_n is None:
                    linked_n = nval
                elif linked_n != nval or nval == 0:
                    raise exceptions.RequestParameterInvalidException(
                        "Failed to match linked batch selections. Please select equal number of data files."
                    )
                linked.append(value["values"])
                linked_keys.append(InputKey(input_id))
        elif isinstance(value, dict) and "batch" in value:
            # remove batch wrapper and render simplified input form rest of workflow
            # code expects
            inputs[input_id] = value["values"][0]

    param_combinations = []
    input_combinations = []
    params_keys = []
    linked = linked or [[None]]
    product = product or [[None]]
    linked_keys = linked_keys or [None]
    product_keys = product_keys or [None]
    for linked_values, product_values in itertools.product(zip(*linked), itertools.product(*product)):
        new_params = copy.deepcopy(param_inputs)
        new_inputs = copy.deepcopy(inputs)
        new_keys = []
        for input_key, value in list(zip(linked_keys, linked_values)) + list(zip(product_keys, product_values)):
            if input_key:
                if isinstance(input_key, ParamKey):
                    step_id = input_key.step_id
                    key = input_key.key
                    assert step_id is not None
                    new_params[step_id][key] = value
                    if "hid" in value:
                        new_keys.append(str(value["hid"]))
                else:
                    input_id = input_key.input_id
                    assert input_id is not None
                    new_inputs[input_id] = value
                    if "hid" in value:
                        new_keys.append(str(value["hid"]))

        params_keys.append(new_keys)
        param_combinations.append(new_params)
        input_combinations.append(new_inputs)

    return WorkflowParameterExpansion(param_combinations, params_keys, input_combinations)


def process_key(incoming_key, incoming_value, d):
    key_parts = incoming_key.split("|")
    if len(key_parts) == 1:
        # Regular parameter
        if incoming_key in d and not incoming_value:
            # In case we get an empty repeat after we already filled in a repeat element
            return
        d[incoming_key] = incoming_value
    elif key_parts[0].rsplit("_", 1)[-1].isdigit():
        # Repeat
        input_name, index = key_parts[0].rsplit("_", 1)
        index = int(index)
        d.setdefault(input_name, [])
        newlist = [{} for _ in range(index + 1)]
        d[input_name].extend(newlist[len(d[input_name]) :])
        subdict = d[input_name][index]
        process_key("|".join(key_parts[1:]), incoming_value=incoming_value, d=subdict)
    else:
        # Section / Conditional
        input_name = key_parts[0]
        subdict = {}
        d[input_name] = subdict
        process_key("|".join(key_parts[1:]), incoming_value=incoming_value, d=subdict)


ExpandedT = Tuple[List[Dict[str, Any]], Optional[matching.MatchingCollections]]


def expand_meta_parameters(trans, tool, incoming) -> ExpandedT:
    """
    Take in a dictionary of raw incoming parameters and expand to a list
    of expanded incoming parameters (one set of parameters per tool
    execution).
    """

    for key in list(incoming.keys()):
        if key.endswith("|__identifier__"):
            incoming.pop(key)

    # If we're going to multiply input dataset combinations
    # order matters, so the following reorders incoming
    # according to tool.inputs (which is ordered).
    incoming_copy = incoming.copy()
    nested_dict: Dict[str, Any] = {}
    for incoming_key, incoming_value in incoming_copy.items():
        if not incoming_key.startswith("__"):
            process_key(incoming_key, incoming_value=incoming_value, d=nested_dict)

    reordered_incoming = {}

    def visitor(input, value, prefix, prefixed_name, prefixed_label, error, **kwargs):
        if prefixed_name in incoming_copy:
            reordered_incoming[prefixed_name] = incoming_copy[prefixed_name]
            del incoming_copy[prefixed_name]

    visit_input_values(inputs=tool.inputs, input_values=nested_dict, callback=visitor)
    reordered_incoming.update(incoming_copy)

    def classifier(input_key):
        value = incoming[input_key]
        if isinstance(value, dict) and "values" in value:
            # Explicit meta wrapper for inputs...
            is_batch = value.get("batch", False)
            is_linked = value.get("linked", True)
            if is_batch and is_linked:
                classification = permutations.input_classification.MATCHED
            elif is_batch:
                classification = permutations.input_classification.MULTIPLIED
            else:
                classification = permutations.input_classification.SINGLE
            if __collection_multirun_parameter(value):
                collection_value = value["values"][0]
                values = __expand_collection_parameter(
                    trans, input_key, collection_value, collections_to_match, linked=is_linked
                )
            else:
                values = value["values"]
        else:
            classification = permutations.input_classification.SINGLE
            values = value
        return classification, values

    collections_to_match = matching.CollectionsToMatch()

    # Stick an unexpanded version of multirun keys so they can be replaced,
    # by expand_mult_inputs.
    incoming_template = reordered_incoming

    expanded_incomings = permutations.expand_multi_inputs(incoming_template, classifier)
    if collections_to_match.has_collections():
        collection_info = trans.app.dataset_collection_manager.match_collections(collections_to_match)
    else:
        collection_info = None
    return expanded_incomings, collection_info


def __expand_collection_parameter(trans, input_key, incoming_val, collections_to_match, linked=False):
    # If subcollectin multirun of data_collection param - value will
    # be "hdca_id|subcollection_type" else it will just be hdca_id
    if "|" in incoming_val:
        encoded_hdc_id, subcollection_type = incoming_val.split("|", 1)
    else:
        try:
            src = incoming_val["src"]
            if src != "hdca":
                raise exceptions.ToolMetaParameterException(f"Invalid dataset collection source type {src}")
            encoded_hdc_id = incoming_val["id"]
            subcollection_type = incoming_val.get("map_over_type", None)
        except TypeError:
            encoded_hdc_id = incoming_val
            subcollection_type = None
    hdc_id = trans.app.security.decode_id(encoded_hdc_id)
    hdc = trans.sa_session.get(HistoryDatasetCollectionAssociation, hdc_id)
    collections_to_match.add(input_key, hdc, subcollection_type=subcollection_type, linked=linked)
    if subcollection_type is not None:
        subcollection_elements = subcollections.split_dataset_collection_instance(hdc, subcollection_type)
        return subcollection_elements
    else:
        hdas = []
        for element in hdc.collection.dataset_elements:
            hda = element.dataset_instance
            hda.element_identifier = element.element_identifier
            hdas.append(hda)
        return hdas


def __collection_multirun_parameter(value):
    is_batch = value.get("batch", False)
    if not is_batch:
        return False

    batch_values = util.listify(value["values"])
    if len(batch_values) == 1:
        batch_over = batch_values[0]
        if isinstance(batch_over, dict) and ("src" in batch_over) and (batch_over["src"] in {"hdca", "dce"}):
            return True
    return False



"""
This module manages loading of Galaxy webhooks.
"""

import logging
import os

import yaml

from galaxy.util import config_directories_from_setting

log = logging.getLogger(__name__)


class Webhook:
    def __init__(self, id, type, activate, weight, path):
        self.id = id
        self.type = type
        self.activate = activate
        self.weight = weight
        self.path = path
        self.styles = ""
        self.script = ""
        self.helper = ""
        self.config = {}

    def to_dict(self):
        return {
            "id": self.id,
            "type": self.type,
            "activate": self.activate,
            "weight": self.weight,
            "styles": self.styles,
            "script": self.script,
            "config": self.config,
        }


class WebhooksRegistry:
    def __init__(self, webhooks_dirs):
        self.webhooks = []
        self.webhooks_directories = []

        for webhook_dir in config_directories_from_setting(webhooks_dirs):
            for plugin_dir in os.listdir(webhook_dir):
                path = os.path.join(webhook_dir, plugin_dir)
                if os.path.isdir(path):
                    self.webhooks_directories.append(path)

        self.load_webhooks()

    def load_webhooks(self):
        for directory in self.webhooks_directories:
            config_file_path = None
            for config_file in ["config.yml", "config.yaml"]:
                path = os.path.join(directory, config_file)
                if os.path.isfile(path):
                    config_file_path = path
                    break

            if config_file_path:
                try:
                    self.load_webhook_from_config(directory, config_file_path)
                except Exception as e:
                    log.exception(e)

    def load_webhook_from_config(self, webhook_dir, config_file_path):
        with open(config_file_path) as fh:
            config = yaml.safe_load(fh)

        weight = config.get("weight", 1)
        if weight < 1:
            raise ValueError("Webhook weight must be greater or equal 1.")

        webhook = Webhook(
            config.get("id"),
            config.get("type"),
            config.get("activate", False),
            weight,
            webhook_dir,
        )

        # Read styles into a string, assuming all styles are in a
        # single file
        try:
            styles_file = os.path.join(webhook_dir, "styles.css")
            with open(styles_file) as fh:
                webhook.styles = fh.read().replace("\n", "")
        except OSError:
            pass

        # Read script into a string, assuming everything is in a
        # single file
        try:
            script_file = os.path.join(webhook_dir, "script.js")
            with open(script_file, encoding="utf-8") as fh:
                webhook.script = fh.read()
        except OSError:
            pass

        # Save helper function path if it exists
        helper_path = os.path.join(webhook_dir, "__init__.py")
        if os.path.isfile(helper_path):
            webhook.helper = helper_path

        webhook.config = config
        self.webhooks.append(webhook)



#!/usr/bin/env python
# Dan Blankenberg
# Selects N random lines from a file and outputs to another file, maintaining original line order
# allows specifying a seed
# does two passes to determine line offsets/count, and then to output contents
from __future__ import print_function

import optparse
import random
from math import (
    ceil,
    log,
)


def randint(a, b):
    return a + int(random.random() * (b + 1 - a))


def sample(population, k):
    """Copied straight from python 2.7."""

    # Sampling without replacement entails tracking either potential
    # selections (the pool) in a list or previous selections in a set.

    # When the number of selections is small compared to the
    # population, then tracking selections is efficient, requiring
    # only a small set and an occasional reselection.  For
    # a larger number of selections, the pool tracking method is
    # preferred since the list takes less space than the
    # set and it doesn't suffer from frequent reselections.

    n = len(population)
    if not 0 <= k <= n:
        raise ValueError("sample larger than population")
    result = [None] * k
    setsize = 21  # size of a small set minus size of an empty list
    if k > 5:
        setsize += 4 ** ceil(log(k * 3, 4))  # table size for big sets
    if n <= setsize or hasattr(population, "keys"):
        # An n-length list is smaller than a k-length set, or this is a
        # mapping type so the other algorithm wouldn't work.
        pool = list(population)
        for i in range(k):  # invariant:  non-selected at [0,n-i)
            j = int(random.random() * (n - i))
            result[i] = pool[j]
            pool[j] = pool[n - i - 1]  # move non-selected item into vacancy
    else:
        try:
            selected = set()
            selected_add = selected.add
            for i in range(k):
                j = int(random.random() * n)
                while j in selected:
                    j = int(random.random() * n)
                selected_add(j)
                result[i] = population[j]
        except (TypeError, KeyError):  # handle (at least) sets
            if isinstance(population, list):
                raise
            return sample(tuple(population), k)
    return result


def get_random_by_subtraction(line_offsets, num_lines):
    while len(line_offsets) > num_lines:
        del line_offsets[randint(0, len(line_offsets) - 1)]
    return line_offsets


def get_random_by_sample(line_offsets, num_lines):
    line_offsets = sample(line_offsets, num_lines)
    line_offsets.sort()
    return line_offsets


def get_random(line_offsets, num_lines):
    if num_lines > (len(line_offsets) / 2):
        return get_random_by_subtraction(line_offsets, num_lines)
    else:
        return get_random_by_sample(line_offsets, num_lines)


def __main__():
    parser = optparse.OptionParser()
    parser.add_option(
        "-s", "--seed", dest="seed", action="store", type="string", default=None, help="Set the random seed."
    )
    (options, args) = parser.parse_args()

    assert len(args) == 3, "Invalid command line specified."

    with open(args[0], "rb") as input, open(args[1], "wb") as output:
        num_lines = int(args[2])
        assert num_lines > 0, "You must select at least one line."

        if options.seed is not None:
            seed_int = int("".join(str(ord(_)) for _ in options.seed))
            try:
                # Select version 1, which results in the same seed for python 2 and 3
                random.seed(seed_int, version=1)
            except TypeError:
                # We're on python 2.X, which doesn't know the version argument
                random.seed(seed_int)

        # get line offsets
        line_offsets = []
        teller = input.tell
        readliner = input.readline
        appender = line_offsets.append
        while True:
            offset = teller()
            if readliner():
                appender(offset)
            else:
                break

        total_lines = len(line_offsets)
        assert num_lines <= total_lines, "Error: asked to select more lines (%i) than there were in the file (%i)." % (
            num_lines,
            total_lines,
        )

        # get random line offsets
        line_offsets = get_random(line_offsets, num_lines)

        # write out random lines
        seeker = input.seek
        writer = output.write
        for line_offset in line_offsets:
            seeker(line_offset)
            writer(readliner())
    print("Kept %i of %i total lines." % (num_lines, total_lines))
    if options.seed is not None:
        print('Used random seed of "%s".' % options.seed)


if __name__ == "__main__":
    __main__()



import logging

from galaxy.model.base import transaction
from . import DefaultToolAction

log = logging.getLogger(__name__)


class DataManagerToolAction(DefaultToolAction):
    """Tool action used for Data Manager Tools"""

    def execute(self, tool, trans, **kwds):
        rval = super().execute(tool, trans, **kwds)
        if isinstance(rval, tuple) and len(rval) >= 2 and isinstance(rval[0], trans.app.model.Job):
            assoc = trans.app.model.DataManagerJobAssociation(job=rval[0], data_manager_id=tool.data_manager_id)
            trans.sa_session.add(assoc)
            with transaction(trans.sa_session):
                trans.sa_session.commit()
        else:
            log.error(f"Got bad return value from DefaultToolAction.execute(): {rval}")
        return rval



import json
import os
import subprocess
from typing import MutableMapping

from cwl_utils.expression import do_eval as _do_eval

from .util import find_engine

FILE_DIRECTORY = os.path.normpath(os.path.dirname(os.path.join(__file__)))
NODE_ENGINE = os.path.join(FILE_DIRECTORY, "cwlNodeEngine.js")


def do_eval(expression: str, context: MutableMapping):
    return _do_eval(
        expression,
        context,
        [{"class": "InlineJavascriptRequirement"}],
        None,
        None,
        {},
        cwlVersion="v1.2.1",
    )


def evaluate(config, input):
    application = find_engine(config)

    default_context = {
        "engineConfig": [],
        "job": {},
        "context": None,
        "outdir": None,
        "tmpdir": None,
    }

    new_input = default_context
    new_input.update(input)

    sp = subprocess.Popen(
        [application, NODE_ENGINE], shell=False, close_fds=True, stdin=subprocess.PIPE, stdout=subprocess.PIPE
    )
    input_str = f"{json.dumps(new_input)}\n\n"
    input_bytes = input_str.encode("utf-8")
    (stdoutdata, stderrdata) = sp.communicate(input_bytes)
    if sp.returncode != 0:
        message = f"Expression engine returned non-zero exit code on evaluation of\n{json.dumps(new_input, indent=4)}{stdoutdata}{stderrdata}"
        raise Exception(message)

    rval_raw = stdoutdata.decode("utf-8")
    return json.loads(rval_raw)



"""
Class encapsulating the management of repositories installed into Galaxy from the Tool Shed.
"""

import copy
import logging
import os
import shutil
from typing import (
    Any,
    Dict,
    List,
    no_type_check,
    Optional,
    Tuple,
)

from galaxy import util
from galaxy.model.base import transaction
from galaxy.model.tool_shed_install import (
    ToolDependency,
    ToolShedRepository,
)
from galaxy.tool_shed.galaxy_install.client import InstallationTarget
from galaxy.tool_shed.galaxy_install.metadata.installed_repository_metadata_manager import (
    InstalledRepositoryMetadataManager,
)
from galaxy.tool_shed.galaxy_install.repository_dependencies import repository_dependency_manager
from galaxy.tool_shed.galaxy_install.tools import (
    data_manager,
    tool_panel_manager,
)
from galaxy.tool_shed.util import (
    repository_util,
    shed_util_common as suc,
    tool_dependency_util,
)
from galaxy.tool_shed.util.container_util import generate_repository_dependencies_key_for_repository
from galaxy.util.tool_shed import common_util
from galaxy.util.tool_shed.xml_util import parse_xml

log = logging.getLogger(__name__)

RepositoryTupleT = Tuple[str, str, str, str]


class InstalledRepositoryManager:
    app: InstallationTarget
    _tool_paths: List[str]
    installed_repository_dicts: List[Dict[str, Any]]
    repository_dependencies_of_installed_repositories: Dict[RepositoryTupleT, List[RepositoryTupleT]]
    installed_repository_dependencies_of_installed_repositories: Dict[RepositoryTupleT, List[RepositoryTupleT]]
    installed_dependent_repositories_of_installed_repositories: Dict[RepositoryTupleT, List[RepositoryTupleT]]

    def __init__(self, app: InstallationTarget):
        """
        Among other things, keep in in-memory sets of tuples defining installed repositories and tool dependencies along with
        the relationships between each of them.  This will allow for quick discovery of those repositories or components that
        can be uninstalled.  The feature allowing a Galaxy administrator to uninstall a repository should not be available to
        repositories or tool dependency packages that are required by other repositories or their contents (packages). The
        uninstall feature should be available only at the repository hierarchy level where every dependency will be uninstalled.
        The exception for this is if an item (repository or tool dependency package) is not in an INSTALLED state - in these
        cases, the specific item can be uninstalled in order to attempt re-installation.
        """
        self.app = app
        self.install_model = self.app.install_model
        self.context = self.install_model.context
        self.tool_configs = self.app.config.tool_configs

        self._tool_paths = []

        self.installed_repository_dicts = []
        # Keep an in-memory dictionary whose keys are tuples defining tool_shed_repository objects (whose status is 'Installed')
        # and whose values are a list of tuples defining tool_shed_repository objects (whose status can be anything) required by
        # the key.  The value defines the entire repository dependency tree.
        self.repository_dependencies_of_installed_repositories = {}
        # Keep an in-memory dictionary whose keys are tuples defining tool_shed_repository objects (whose status is 'Installed')
        # and whose values are a list of tuples defining tool_shed_repository objects (whose status is 'Installed') required by
        # the key.  The value defines the entire repository dependency tree.
        self.installed_repository_dependencies_of_installed_repositories = {}
        # Keep an in-memory dictionary whose keys are tuples defining tool_shed_repository objects (whose status is 'Installed')
        # and whose values are a list of tuples defining tool_shed_repository objects (whose status is 'Installed') that require
        # the key.
        self.installed_dependent_repositories_of_installed_repositories = {}

    @property
    def tool_paths(self) -> List[str]:
        """Return all possible tool_path attributes of all tool config files."""
        if len(self._tool_paths) != len(self.tool_configs):
            # This could be happen at startup or after the creation of a new shed_tool_conf.xml file
            # before the installation of the first repository
            tool_paths = []
            for tool_config in self.tool_configs:
                tree, error_message = parse_xml(tool_config)
                if error_message:
                    log.error(error_message)
                else:
                    assert tree
                    tool_path = tree.getroot().get("tool_path")
                    if tool_path:
                        tool_paths.append(tool_path)
            self._tool_paths = tool_paths
        return self._tool_paths

    def activate_repository(self, repository: ToolShedRepository) -> None:
        """Activate an installed tool shed repository that has been marked as deactivated."""
        shed_tool_conf, tool_path, relative_install_dir = suc.get_tool_panel_config_tool_path_install_dir(
            self.app, repository
        )
        repository.deleted = False
        repository.status = ToolShedRepository.installation_status.INSTALLED
        if repository.includes_tools_for_display_in_tool_panel:
            repository_clone_url = common_util.generate_clone_url_for_installed_repository(self.app, repository)
            tpm = tool_panel_manager.ToolPanelManager(self.app)
            irmm = InstalledRepositoryMetadataManager(
                app=self.app,
                tpm=tpm,
                repository=repository,
                changeset_revision=repository.changeset_revision,
                metadata_dict=repository.metadata_,
            )
            repository_tools_tups = irmm.get_repository_tools_tups()
            # Reload tools into the appropriate tool panel section.
            tool_panel_dict = repository.metadata_["tool_panel_section"]
            tpm.add_to_tool_panel(
                repository.name,
                repository_clone_url,
                repository.installed_changeset_revision,
                repository_tools_tups,
                repository.owner,
                shed_tool_conf,
                tool_panel_dict,
                new_install=False,
            )
            if repository.includes_data_managers:
                tp, data_manager_relative_install_dir = repository.get_tool_relative_path(self.app)
                # Hack to add repository.name here, which is actually the root of the installed repository
                data_manager_relative_install_dir = os.path.join(data_manager_relative_install_dir, repository.name)
                dmh = data_manager.DataManagerHandler(self.app)
                dmh.install_data_managers(
                    self.app.config.shed_data_manager_config_file,
                    repository.metadata_,
                    repository.get_shed_config_dict(self.app),
                    data_manager_relative_install_dir,
                    repository,
                    repository_tools_tups,
                )
        self.context.add(repository)
        with transaction(self.context):
            self.context.commit()

    def add_entry_to_installed_repository_dependencies_of_installed_repositories(
        self, repository: ToolShedRepository
    ) -> None:
        """
        Add an entry to self.installed_repository_dependencies_of_installed_repositories.  A side-effect of this method
        is the population of self.installed_dependent_repositories_of_installed_repositories.  Since this method discovers
        all repositories required by the received repository, it can use the list to add entries to the reverse dictionary.
        """
        repository_tup = self.get_repository_tuple_for_installed_repository_manager(repository)
        tool_shed, name, owner, installed_changeset_revision = repository_tup
        # Get the list of repository dependencies for this repository.
        status = ToolShedRepository.installation_status.INSTALLED
        repository_dependency_tups = self.get_repository_dependency_tups_for_installed_repository(
            repository, status=status
        )
        # Add an entry to self.installed_repository_dependencies_of_installed_repositories.
        if repository_tup not in self.installed_repository_dependencies_of_installed_repositories:
            debug_msg = (
                f"Adding an entry for revision {installed_changeset_revision} of repository {name} owned by {owner} "
            )
            debug_msg += "to installed_repository_dependencies_of_installed_repositories."
            log.debug(debug_msg)
            self.installed_repository_dependencies_of_installed_repositories[repository_tup] = (
                repository_dependency_tups
            )
        # Use the repository_dependency_tups to add entries to the reverse dictionary
        # self.installed_dependent_repositories_of_installed_repositories.
        for required_repository_tup in repository_dependency_tups:
            debug_msg = f"Appending revision {installed_changeset_revision} of repository {name} owned by {owner} "
            debug_msg += "to all dependent repositories in installed_dependent_repositories_of_installed_repositories."
            log.debug(debug_msg)
            if required_repository_tup in self.installed_dependent_repositories_of_installed_repositories:
                self.installed_dependent_repositories_of_installed_repositories[required_repository_tup].append(
                    repository_tup
                )
            else:
                self.installed_dependent_repositories_of_installed_repositories[required_repository_tup] = [
                    repository_tup
                ]

    def add_entry_to_repository_dependencies_of_installed_repositories(self, repository: ToolShedRepository) -> None:
        """Add an entry to self.repository_dependencies_of_installed_repositories."""
        repository_tup = self.get_repository_tuple_for_installed_repository_manager(repository)
        if repository_tup not in self.repository_dependencies_of_installed_repositories:
            tool_shed, name, owner, installed_changeset_revision = repository_tup
            debug_msg = (
                f"Adding an entry for revision {installed_changeset_revision} of repository {name} owned by {owner} "
            )
            debug_msg += "to repository_dependencies_of_installed_repositories."
            log.debug(debug_msg)
            repository_dependency_tups = self.get_repository_dependency_tups_for_installed_repository(
                repository, status=None
            )
            self.repository_dependencies_of_installed_repositories[repository_tup] = repository_dependency_tups

    def get_containing_repository_for_tool_dependency(self, tool_dependency_tup: tuple) -> ToolShedRepository:
        tool_shed_repository_id, name, version, type = tool_dependency_tup
        return self.context.query(ToolShedRepository).get(tool_shed_repository_id)

    def get_dependencies_for_repository(
        self,
        tool_shed_url: str,
        repo_info_dict: Dict[str, repository_util.AnyRepositoryTupleT],
        includes_tool_dependencies,
        updating=False,
    ) -> Dict[str, Any]:
        """
        Return dictionaries containing the sets of installed and missing tool dependencies and repository
        dependencies associated with the repository defined by the received repo_info_dict.
        """
        rdim = repository_dependency_manager.RepositoryDependencyInstallManager(self.app)
        repository = None
        installed_rd: Dict[str, Any] = {}
        installed_td: repository_util.ToolDependenciesDictT = {}
        missing_rd: Dict[str, Any] = {}
        missing_td: repository_util.ToolDependenciesDictT = {}
        name = next(iter(repo_info_dict))
        repo_info_tuple = repo_info_dict[name]
        (
            description,
            repository_clone_url,
            changeset_revision,
            ctx_rev,
            repository_owner,
            repository_dependencies,
            tool_dependencies,
        ) = repository_util.get_repo_info_tuple_contents(repo_info_tuple)
        if tool_dependencies:
            if not includes_tool_dependencies:
                includes_tool_dependencies = True
            # Inspect the tool_dependencies dictionary to separate the installed and missing tool dependencies.
            # We don't add to installed_td and missing_td here because at this point they are empty.
            installed_td, missing_td = self.get_installed_and_missing_tool_dependencies_for_repository(
                tool_dependencies
            )
        # In cases where a repository dependency is required only for compiling a dependent repository's
        # tool dependency, the value of repository_dependencies will be an empty dictionary here.
        if repository_dependencies:
            # We have a repository with one or more defined repository dependencies.
            if not repository:
                repository = repository_util.get_repository_for_dependency_relationship(
                    self.app, tool_shed_url, name, repository_owner, changeset_revision
                )
            if not updating and repository and repository.metadata_:
                installed_rd, missing_rd = self.get_installed_and_missing_repository_dependencies(repository)
            else:
                (
                    installed_rd,
                    missing_rd,
                ) = self.get_installed_and_missing_repository_dependencies_for_new_or_updated_install(repo_info_tuple)
            # Discover all repository dependencies and retrieve information for installing them.
            all_repo_info_dict = rdim.get_required_repo_info_dicts(tool_shed_url, util.listify(repo_info_dict))
            has_repository_dependencies = all_repo_info_dict.get("has_repository_dependencies", False)
            has_repository_dependencies_only_if_compiling_contained_td = all_repo_info_dict.get(
                "has_repository_dependencies_only_if_compiling_contained_td", False
            )
            includes_tools_for_display_in_tool_panel = all_repo_info_dict.get(
                "includes_tools_for_display_in_tool_panel", False
            )
            includes_tool_dependencies = all_repo_info_dict.get("includes_tool_dependencies", False)
            includes_tools = all_repo_info_dict.get("includes_tools", False)
            required_repo_info_dicts = all_repo_info_dict.get("all_repo_info_dicts", [])
            # Display tool dependencies defined for each of the repository dependencies.
            if required_repo_info_dicts:
                required_tool_dependencies = {}
                for rid in required_repo_info_dicts:
                    for repo_info_tuple in rid.values():
                        (
                            description,
                            repository_clone_url,
                            changeset_revision,
                            ctx_rev,
                            repository_owner,
                            rid_repository_dependencies,
                            rid_tool_dependencies,
                        ) = repository_util.get_repo_info_tuple_contents(repo_info_tuple)
                        if rid_tool_dependencies:
                            for td_key, td_dict in rid_tool_dependencies.items():
                                if td_key not in required_tool_dependencies:
                                    required_tool_dependencies[td_key] = td_dict
                if required_tool_dependencies:
                    # Discover and categorize all tool dependencies defined for this repository's repository dependencies.
                    (
                        required_installed_td,
                        required_missing_td,
                    ) = self.get_installed_and_missing_tool_dependencies_for_repository(required_tool_dependencies)
                    if required_installed_td:
                        if not includes_tool_dependencies:
                            includes_tool_dependencies = True
                        for td_key, td_dict in required_installed_td.items():
                            if td_key not in installed_td:
                                installed_td[td_key] = td_dict
                    if required_missing_td:
                        if not includes_tool_dependencies:
                            includes_tool_dependencies = True
                        for td_key, td_dict in required_missing_td.items():
                            if td_key not in missing_td:
                                missing_td[td_key] = td_dict
        else:
            # We have a single repository with (possibly) no defined repository dependencies.
            all_repo_info_dict = rdim.get_required_repo_info_dicts(tool_shed_url, util.listify(repo_info_dict))
            has_repository_dependencies = all_repo_info_dict.get("has_repository_dependencies", False)
            has_repository_dependencies_only_if_compiling_contained_td = all_repo_info_dict.get(
                "has_repository_dependencies_only_if_compiling_contained_td", False
            )
            includes_tools_for_display_in_tool_panel = all_repo_info_dict.get(
                "includes_tools_for_display_in_tool_panel", False
            )
            includes_tool_dependencies = all_repo_info_dict.get("includes_tool_dependencies", False)
            includes_tools = all_repo_info_dict.get("includes_tools", False)
            required_repo_info_dicts = all_repo_info_dict.get("all_repo_info_dicts", [])
        dependencies_for_repository_dict = dict(
            changeset_revision=changeset_revision,
            has_repository_dependencies=has_repository_dependencies,
            has_repository_dependencies_only_if_compiling_contained_td=has_repository_dependencies_only_if_compiling_contained_td,
            includes_tool_dependencies=includes_tool_dependencies,
            includes_tools=includes_tools,
            includes_tools_for_display_in_tool_panel=includes_tools_for_display_in_tool_panel,
            installed_repository_dependencies=installed_rd,
            installed_tool_dependencies=installed_td,
            missing_repository_dependencies=missing_rd,
            missing_tool_dependencies=missing_td,
            name=name,
            repository_owner=repository_owner,
        )
        return dependencies_for_repository_dict

    def get_installed_and_missing_repository_dependencies(
        self, repository: ToolShedRepository
    ) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        """
        Return the installed and missing repository dependencies for a tool shed repository that has a record
        in the Galaxy database, but may or may not be installed.  In this case, the repository dependencies are
        associated with the repository in the database.  Do not include a repository dependency if it is required
        only to compile a tool dependency defined for the dependent repository since these special kinds of repository
        dependencies are really a dependency of the dependent repository's contained tool dependency, and only
        if that tool dependency requires compilation.
        """
        missing_repository_dependencies: Dict[str, Any] = {}
        installed_repository_dependencies: Dict[str, Any] = {}
        has_repository_dependencies = repository.has_repository_dependencies
        if has_repository_dependencies:
            # The repository dependencies container will include only the immediate repository
            # dependencies of this repository, so the container will be only a single level in depth.
            metadata = repository.metadata_
            installed_rd_tups = []
            missing_rd_tups = []
            for tsr in repository.repository_dependencies:
                prior_installation_required = self.set_prior_installation_required(repository, tsr)
                only_if_compiling_contained_td = self.set_only_if_compiling_contained_td(repository, tsr)
                rd_tup = [
                    tsr.tool_shed,
                    tsr.name,
                    tsr.owner,
                    tsr.changeset_revision,
                    prior_installation_required,
                    only_if_compiling_contained_td,
                    tsr.id,
                    tsr.status,
                ]
                if tsr.status == ToolShedRepository.installation_status.INSTALLED:
                    installed_rd_tups.append(rd_tup)
                else:
                    # We'll only add the rd_tup to the missing_rd_tups list if the received repository
                    # has tool dependencies that are not correctly installed.  This may prove to be a
                    # weak check since the repository in question may not have anything to do with
                    # compiling the missing tool dependencies.  If we discover that this is a problem,
                    # more granular checking will be necessary here.
                    if repository.missing_tool_dependencies:
                        if not self.repository_dependency_needed_only_for_compiling_tool_dependency(repository, tsr):
                            missing_rd_tups.append(rd_tup)
                    else:
                        missing_rd_tups.append(rd_tup)
            if installed_rd_tups or missing_rd_tups:
                # Get the description from the metadata in case it has a value.
                repository_dependencies = metadata.get("repository_dependencies", {})
                description = repository_dependencies.get("description", None)
                # We need to add a root_key entry to one or both of installed_repository_dependencies dictionary and the
                # missing_repository_dependencies dictionaries for proper display parsing.
                root_key = generate_repository_dependencies_key_for_repository(
                    repository.tool_shed,
                    repository.name,
                    repository.owner,
                    repository.installed_changeset_revision,
                    prior_installation_required,
                    only_if_compiling_contained_td,
                )
                if installed_rd_tups:
                    installed_repository_dependencies["root_key"] = root_key
                    installed_repository_dependencies[root_key] = installed_rd_tups
                    installed_repository_dependencies["description"] = description
                if missing_rd_tups:
                    missing_repository_dependencies["root_key"] = root_key
                    missing_repository_dependencies[root_key] = missing_rd_tups
                    missing_repository_dependencies["description"] = description
        return installed_repository_dependencies, missing_repository_dependencies

    def get_installed_and_missing_repository_dependencies_for_new_or_updated_install(
        self, repo_info_tuple
    ) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        """
        Parse the received repository_dependencies dictionary that is associated with a repository being
        installed into Galaxy for the first time and attempt to determine repository dependencies that are
        already installed and those that are not.
        """
        missing_repository_dependencies = {}
        installed_repository_dependencies = {}
        missing_rd_tups = []
        installed_rd_tups = []
        (
            description,
            repository_clone_url,
            changeset_revision,
            ctx_rev,
            repository_owner,
            repository_dependencies,
            tool_dependencies,
        ) = repository_util.get_repo_info_tuple_contents(repo_info_tuple)
        if repository_dependencies:
            description = repository_dependencies["description"]
            root_key = repository_dependencies["root_key"]
            # The repository dependencies container will include only the immediate repository dependencies of
            # this repository, so the container will be only a single level in depth.
            for key, rd_tups in repository_dependencies.items():
                if key in ["description", "root_key"]:
                    continue
                for rd_tup in rd_tups:
                    (
                        tool_shed,
                        name,
                        owner,
                        changeset_revision,
                        prior_installation_required,
                        only_if_compiling_contained_td,
                    ) = common_util.parse_repository_dependency_tuple(rd_tup)
                    # Updates to installed repository revisions may have occurred, so make sure to locate the
                    # appropriate repository revision if one exists.  We need to create a temporary repo_info_tuple
                    # that includes the correct repository owner which we get from the current rd_tup.  The current
                    # tuple looks like: ( description, repository_clone_url, changeset_revision, ctx_rev, repository_owner,
                    #                     repository_dependencies, installed_td )
                    tmp_clone_url = common_util.generate_clone_url_from_repo_info_tup(self.app, rd_tup)
                    tmp_repo_info_tuple = (None, tmp_clone_url, changeset_revision, None, owner, None, None)
                    repository, installed_changeset_revision = repository_util.repository_was_previously_installed(
                        self.app, tool_shed, name, tmp_repo_info_tuple, from_tip=False
                    )
                    if repository:
                        new_rd_tup = [
                            tool_shed,
                            name,
                            owner,
                            changeset_revision,
                            prior_installation_required,
                            only_if_compiling_contained_td,
                            repository.id,
                            repository.status,
                        ]
                        if repository.status == ToolShedRepository.installation_status.INSTALLED:
                            if new_rd_tup not in installed_rd_tups:
                                installed_rd_tups.append(new_rd_tup)
                        else:
                            # A repository dependency that is not installed will not be considered missing if its value
                            # for only_if_compiling_contained_td is True  This is because this type of repository dependency
                            # will only be considered at the time that the specified tool dependency is being installed, and
                            # even then only if the compiled binary of the tool dependency could not be installed due to the
                            # unsupported installation environment.
                            if not util.asbool(only_if_compiling_contained_td):
                                if new_rd_tup not in missing_rd_tups:
                                    missing_rd_tups.append(new_rd_tup)
                    else:
                        new_rd_tup = [
                            tool_shed,
                            name,
                            owner,
                            changeset_revision,
                            prior_installation_required,
                            only_if_compiling_contained_td,
                            None,
                            "Never installed",
                        ]
                        if not util.asbool(only_if_compiling_contained_td):
                            # A repository dependency that is not installed will not be considered missing if its value for
                            # only_if_compiling_contained_td is True - see above...
                            if new_rd_tup not in missing_rd_tups:
                                missing_rd_tups.append(new_rd_tup)
        if installed_rd_tups:
            installed_repository_dependencies["root_key"] = root_key
            installed_repository_dependencies[root_key] = installed_rd_tups
            installed_repository_dependencies["description"] = description
        if missing_rd_tups:
            missing_repository_dependencies["root_key"] = root_key
            missing_repository_dependencies[root_key] = missing_rd_tups
            missing_repository_dependencies["description"] = description
        return installed_repository_dependencies, missing_repository_dependencies

    # The following function will be removed at some point and has clear issues the type checking
    # makes clear... I'm going to skip type checking for now rather than fix bugs in deprecated code
    @no_type_check
    def get_installed_and_missing_tool_dependencies_for_repository(
        self, tool_dependencies_dict: repository_util.ToolDependenciesDictT
    ) -> Tuple[repository_util.ToolDependenciesDictT, repository_util.ToolDependenciesDictT]:
        """
        Return the lists of installed tool dependencies and missing tool dependencies for a set of repositories
        being installed into Galaxy.
        """
        # FIXME: This implementation breaks when updates to a repository contain dependencies that result in
        # multiple entries for a specific tool dependency.  A scenario where this can happen is where 2 repositories
        # define  the same dependency internally (not using the complex repository dependency definition to a separate
        # package repository approach).  If 2 repositories contain the same tool_dependencies.xml file, one dependency
        # will be lost since the values in these returned dictionaries are not lists.  All tool dependency dictionaries
        # should have lists as values.  These scenarios are probably extreme corner cases, but still should be handled.
        installed_tool_dependencies: dict = {}
        missing_tool_dependencies: dict = {}
        if tool_dependencies_dict:
            # Make sure not to change anything in the received tool_dependencies_dict as that would be a bad side-effect!
            tmp_tool_dependencies_dict: dict = copy.deepcopy(tool_dependencies_dict)
            for td_key, val in tmp_tool_dependencies_dict.items():
                # Default the status to NEVER_INSTALLED.
                tool_dependency_status = ToolDependency.installation_status.NEVER_INSTALLED
                if td_key == "set_environment":
                    # Set environment tool dependencies are a list.
                    assert isinstance(val, list)
                    new_val = []
                    for requirement_dict in val:
                        # {'repository_name': 'xx',
                        #  'name': 'bwa',
                        #  'version': '0.5.9',
                        #  'repository_owner': 'yy',
                        #  'changeset_revision': 'zz',
                        #  'type': 'package'}
                        tool_dependency = tool_dependency_util.get_tool_dependency_by_name_version_type(
                            self.app,
                            requirement_dict.get("name", None),
                            requirement_dict.get("version", None),
                            requirement_dict.get("type", "package"),
                        )
                        if tool_dependency:
                            tool_dependency_status = tool_dependency.status
                        requirement_dict["status"] = tool_dependency_status
                        new_val.append(requirement_dict)
                        if tool_dependency_status in [ToolDependency.installation_status.INSTALLED]:
                            if td_key in installed_tool_dependencies:
                                installed_tool_dependencies[td_key].extend(new_val)
                            else:
                                installed_tool_dependencies[td_key] = new_val
                        else:
                            if td_key in missing_tool_dependencies:
                                missing_tool_dependencies[td_key].extend(new_val)
                            else:
                                missing_tool_dependencies[td_key] = new_val
                else:
                    # The val dictionary looks something like this:
                    # {'repository_name': 'xx',
                    #  'name': 'bwa',
                    #  'version': '0.5.9',
                    #  'repository_owner': 'yy',
                    #  'changeset_revision': 'zz',
                    #  'type': 'package'}
                    tool_dependency = tool_dependency_util.get_tool_dependency_by_name_version_type(
                        self.app, val.get("name", None), val.get("version", None), val.get("type", "package")
                    )
                    if tool_dependency:
                        tool_dependency_status = tool_dependency.status
                    val["status"] = tool_dependency_status
                if tool_dependency_status in [ToolDependency.installation_status.INSTALLED]:
                    installed_tool_dependencies[td_key] = val
                else:
                    missing_tool_dependencies[td_key] = val
        return installed_tool_dependencies, missing_tool_dependencies

    def get_repository_dependency_tups_for_installed_repository(
        self, repository, dependency_tups=None, status=None
    ) -> List[RepositoryTupleT]:
        """
        Return a list of of tuples defining tool_shed_repository objects (whose status can be anything) required by the
        received repository.  The returned list defines the entire repository dependency tree.  This method is called
        only from Galaxy.
        """
        if dependency_tups is None:
            dependency_tups = []
        repository_tup = self.get_repository_tuple_for_installed_repository_manager(repository)
        for rrda in repository.required_repositories:
            repository_dependency = rrda.repository_dependency
            required_repository = repository_dependency.repository
            if status is None or required_repository.status == status:
                required_repository_tup = self.get_repository_tuple_for_installed_repository_manager(
                    required_repository
                )
                if required_repository_tup == repository_tup:
                    # We have a circular repository dependency relationship, skip this entry.
                    continue
                if required_repository_tup not in dependency_tups:
                    dependency_tups.append(required_repository_tup)
                    return self.get_repository_dependency_tups_for_installed_repository(
                        required_repository, dependency_tups=dependency_tups
                    )
        return dependency_tups

    def get_repository_tuple_for_installed_repository_manager(self, repository: ToolShedRepository) -> RepositoryTupleT:
        return (
            str(repository.tool_shed),
            str(repository.name),
            str(repository.owner),
            str(repository.installed_changeset_revision),
        )

    def get_repository_install_dir(self, tool_shed_repository: ToolShedRepository) -> Optional[str]:
        for tool_path in self.tool_paths:
            ts = common_util.remove_port_from_tool_shed_url(str(tool_shed_repository.tool_shed))
            relative_path = os.path.join(
                tool_path,
                ts,
                "repos",
                str(tool_shed_repository.owner),
                str(tool_shed_repository.name),
                str(tool_shed_repository.installed_changeset_revision),
            )
            if os.path.exists(relative_path):
                return relative_path
        return None

    def handle_existing_tool_dependencies_that_changed_in_update(
        self, repository: ToolShedRepository, original_dependency_dict, new_dependency_dict
    ) -> Tuple[List[str], List[str]]:
        """
        This method is called when a Galaxy admin is getting updates for an installed tool shed
        repository in order to cover the case where an existing tool dependency was changed (e.g.,
        the version of the dependency was changed) but the tool version for which it is a dependency
        was not changed.  In this case, we only want to determine if any of the dependency information
        defined in original_dependency_dict was changed in new_dependency_dict.  We don't care if new
        dependencies were added in new_dependency_dict since they will just be treated as missing
        dependencies for the tool.
        """
        updated_tool_dependency_names = []
        deleted_tool_dependency_names = []
        for original_dependency_key, original_dependency_val_dict in original_dependency_dict.items():
            if original_dependency_key not in new_dependency_dict:
                updated_tool_dependency = self._update_existing_tool_dependency(
                    repository, original_dependency_val_dict, new_dependency_dict
                )
                if updated_tool_dependency:
                    updated_tool_dependency_names.append(updated_tool_dependency.name)
                else:
                    deleted_tool_dependency_names.append(original_dependency_val_dict["name"])
        return updated_tool_dependency_names, deleted_tool_dependency_names

    def uninstall_repository(self, repository: ToolShedRepository, remove_from_disk=True) -> str:
        errors = ""
        shed_tool_conf, tool_path, relative_install_dir = suc.get_tool_panel_config_tool_path_install_dir(
            app=self.app, repository=repository
        )
        if relative_install_dir:
            if tool_path:
                relative_install_dir = os.path.join(tool_path, relative_install_dir)
            repository_install_dir = os.path.abspath(relative_install_dir)
        else:
            repository_install_dir = None
        if repository.includes_tools_for_display_in_tool_panel:
            # Handle tool panel alterations.
            tpm = tool_panel_manager.ToolPanelManager(app=self.app)
            tpm.remove_repository_contents(repository, shed_tool_conf, uninstall=remove_from_disk)
        if repository.includes_data_managers:
            dmh = data_manager.DataManagerHandler(app=self.app)
            dmh.remove_from_data_manager(repository)
        if remove_from_disk:
            try:
                # Remove the repository from disk.
                shutil.rmtree(repository_install_dir)
                log.debug(f"Removed repository installation directory: {repository_install_dir}")
                removed = True
            except Exception as e:
                log.debug(f"Error removing repository installation directory {repository_install_dir}: {e}")
                if isinstance(e, OSError) and not os.path.exists(repository_install_dir):
                    removed = True
                    log.debug("Repository directory does not exist on disk, marking as uninstalled.")
                else:
                    removed = False
            if removed:
                repository.uninstalled = True
                # Remove all installed tool dependencies and tool dependencies stuck in the INSTALLING state, but don't touch any
                # repository dependencies.
                tool_dependencies_to_uninstall = repository.tool_dependencies_installed_or_in_error
                tool_dependencies_to_uninstall.extend(repository.tool_dependencies_being_installed)
                for tool_dependency in tool_dependencies_to_uninstall:
                    uninstalled, error_message = tool_dependency_util.remove_tool_dependency(self.app, tool_dependency)
                    if error_message:
                        errors = f"{errors}  {error_message}"
        repository.deleted = True
        if remove_from_disk:
            repository.status = ToolShedRepository.installation_status.UNINSTALLED
            repository.error_message = None
        else:
            repository.status = ToolShedRepository.installation_status.DEACTIVATED
        self.context.add(repository)
        with transaction(self.context):
            self.context.commit()
        return errors

    def remove_entry_from_installed_repository_dependencies_of_installed_repositories(
        self, repository: ToolShedRepository
    ) -> None:
        """
        Remove an entry from self.installed_repository_dependencies_of_installed_repositories.  A side-effect of this method
        is removal of appropriate value items from self.installed_dependent_repositories_of_installed_repositories.
        """
        # Remove tuples defining this repository from value lists in self.installed_dependent_repositories_of_installed_repositories.
        repository_tup = self.get_repository_tuple_for_installed_repository_manager(repository)
        tool_shed, name, owner, installed_changeset_revision = repository_tup
        altered_installed_dependent_repositories_of_installed_repositories = {}
        for r_tup, v_tups in self.installed_dependent_repositories_of_installed_repositories.items():
            if repository_tup in v_tups:
                debug_msg = (
                    f"Removing entry for revision {installed_changeset_revision} of repository {name} owned by {owner} "
                )
                r_tool_shed, r_name, r_owner, r_installed_changeset_revision = r_tup
                debug_msg += f"from the dependent list for revision {r_installed_changeset_revision} of repository {r_name} owned by {r_owner} "
                debug_msg += "in installed_repository_dependencies_of_installed_repositories."
                log.debug(debug_msg)
                v_tups.remove(repository_tup)
            altered_installed_dependent_repositories_of_installed_repositories[r_tup] = v_tups
        self.installed_dependent_repositories_of_installed_repositories = (
            altered_installed_dependent_repositories_of_installed_repositories
        )
        # Remove this repository's entry from self.installed_repository_dependencies_of_installed_repositories.
        if repository_tup in self.installed_repository_dependencies_of_installed_repositories:
            debug_msg = (
                f"Removing entry for revision {installed_changeset_revision} of repository {name} owned by {owner} "
            )
            debug_msg += "from installed_repository_dependencies_of_installed_repositories."
            log.debug(debug_msg)
            del self.installed_repository_dependencies_of_installed_repositories[repository_tup]

    def remove_entry_from_repository_dependencies_of_installed_repositories(
        self, repository: ToolShedRepository
    ) -> None:
        """Remove an entry from self.repository_dependencies_of_installed_repositories."""
        repository_tup = self.get_repository_tuple_for_installed_repository_manager(repository)
        if repository_tup in self.repository_dependencies_of_installed_repositories:
            tool_shed, name, owner, installed_changeset_revision = repository_tup
            debug_msg = (
                f"Removing entry for revision {installed_changeset_revision} of repository {name} owned by {owner} "
            )
            debug_msg += "from repository_dependencies_of_installed_repositories."
            log.debug(debug_msg)
            del self.repository_dependencies_of_installed_repositories[repository_tup]

    def repository_dependency_needed_only_for_compiling_tool_dependency(
        self, repository: ToolShedRepository, repository_dependency
    ) -> bool:
        for rd_tup in repository.tuples_of_repository_dependencies_needed_for_compiling_td:
            (
                tool_shed,
                name,
                owner,
                changeset_revision,
                prior_installation_required,
                only_if_compiling_contained_td,
            ) = rd_tup
            # TODO: we may discover that we need to check more than just installed_changeset_revision and changeset_revision here, in which
            # case we'll need to contact the tool shed to get the list of all possible changeset_revisions.
            cleaned_tool_shed = common_util.remove_protocol_and_port_from_tool_shed_url(tool_shed)
            cleaned_repository_dependency_tool_shed = common_util.remove_protocol_and_port_from_tool_shed_url(
                str(repository_dependency.tool_shed)
            )
            if (
                cleaned_repository_dependency_tool_shed == cleaned_tool_shed
                and repository_dependency.name == name
                and repository_dependency.owner == owner
                and (
                    repository_dependency.installed_changeset_revision == changeset_revision
                    or repository_dependency.changeset_revision == changeset_revision
                )
            ):
                return True
        return False

    def set_only_if_compiling_contained_td(self, repository, required_repository):
        """
        Return True if the received required_repository is only needed to compile a tool
        dependency defined for the received repository.
        """
        # This method is called only from Galaxy when rendering repository dependencies
        # for an installed tool shed repository.
        # TODO: Do we need to check more than changeset_revision here?
        required_repository_tup = [
            required_repository.tool_shed,
            required_repository.name,
            required_repository.owner,
            required_repository.changeset_revision,
        ]
        for tup in repository.tuples_of_repository_dependencies_needed_for_compiling_td:
            partial_tup = tup[0:4]
            if partial_tup == required_repository_tup:
                return "True"
        return "False"

    def set_prior_installation_required(self, repository, required_repository) -> str:
        """
        Return True if the received required_repository must be installed before the
        received repository.
        """
        tool_shed_url = common_util.get_tool_shed_url_from_tool_shed_registry(
            self.app, str(required_repository.tool_shed)
        )
        required_repository_tup = [
            tool_shed_url,
            str(required_repository.name),
            str(required_repository.owner),
            str(required_repository.changeset_revision),
        ]
        # Get the list of repository dependency tuples associated with the received repository
        # where prior_installation_required is True.
        required_rd_tups_that_must_be_installed = repository.requires_prior_installation_of
        for required_rd_tup in required_rd_tups_that_must_be_installed:
            # Repository dependency tuples in metadata include a prior_installation_required value,
            # so strip it for comparision.
            partial_required_rd_tup = required_rd_tup[0:4]
            if partial_required_rd_tup == required_repository_tup:
                # Return the string value of prior_installation_required, which defaults to 'False'.
                return str(required_rd_tup[4])
        return "False"

    def _update_existing_tool_dependency(
        self, repository: ToolShedRepository, original_dependency_dict, new_dependencies_dict
    ):
        """
        Update an exsiting tool dependency whose definition was updated in a change set
        pulled by a Galaxy administrator when getting updates to an installed tool shed
        repository.  The original_dependency_dict is a single tool dependency definition,
        an example of which is::

            {"name": "bwa",
             "readme": "\\nCompiling BWA requires zlib and libpthread to be present on your system.\\n        ",
             "type": "package",
             "version": "0.6.2"}

        The new_dependencies_dict is the dictionary generated by the metadata_util.generate_tool_dependency_metadata method.
        """
        new_tool_dependency = None
        original_name = original_dependency_dict["name"]
        original_type = original_dependency_dict["type"]
        original_version = original_dependency_dict["version"]
        # Locate the appropriate tool_dependency associated with the repository.
        tool_dependency = None
        for tool_dependency in repository.tool_dependencies:
            if (
                tool_dependency.name == original_name
                and tool_dependency.type == original_type
                and tool_dependency.version == original_version
            ):
                break
        if tool_dependency and tool_dependency.can_update:
            dependency_install_dir = tool_dependency.installation_directory(self.app)
            removed_from_disk, error_message = tool_dependency_util.remove_tool_dependency_installation_directory(
                dependency_install_dir
            )
            if removed_from_disk:
                context = self.context
                new_dependency_name = None
                new_dependency_type = None
                new_dependency_version = None
                for new_dependency_val_dict in new_dependencies_dict.values():
                    # Match on name only, hopefully this will be enough!
                    if original_name == new_dependency_val_dict["name"]:
                        new_dependency_name = new_dependency_val_dict["name"]
                        new_dependency_type = new_dependency_val_dict["type"]
                        new_dependency_version = new_dependency_val_dict["version"]
                        break
                if new_dependency_name and new_dependency_type and new_dependency_version:
                    # Update all attributes of the tool_dependency record in the database.
                    log.debug(
                        "Updating version %s of tool dependency %s %s to have new version %s and type %s.",
                        tool_dependency.version,
                        tool_dependency.type,
                        tool_dependency.name,
                        new_dependency_version,
                        new_dependency_type,
                    )
                    tool_dependency.type = new_dependency_type
                    tool_dependency.version = new_dependency_version
                    tool_dependency.status = ToolDependency.installation_status.UNINSTALLED
                    tool_dependency.error_message = None
                    context.add(tool_dependency)
                    with transaction(context):
                        context.commit()
                    new_tool_dependency = tool_dependency
                else:
                    # We have no new tool dependency definition based on a matching dependency name, so remove
                    # the existing tool dependency record from the database.
                    log.debug(
                        "Deleting version %s of tool dependency %s %s from the database since it is no longer defined.",
                        tool_dependency.version,
                        tool_dependency.type,
                        tool_dependency.name,
                    )
                    context.delete(tool_dependency)
                    with transaction(context):
                        context.commit()
        return new_tool_dependency



"""Typed description of Galaxy's app object."""

import abc
from typing import (
    Any,
    Optional,
    TYPE_CHECKING,
)

from kombu import Connection
from typing_extensions import Protocol

from galaxy.auth import AuthManager
from galaxy.datatypes.registry import Registry
from galaxy.di import Container
from galaxy.files import ConfiguredFileSources
from galaxy.job_metrics import JobMetrics
from galaxy.managers.dbkeys import GenomeBuilds
from galaxy.model.base import (
    ModelMapping,
    SharedModelMapping,
)
from galaxy.model.mapping import GalaxyModelMapping
from galaxy.model.security import (
    GalaxyRBACAgent,
    HostAgent,
)
from galaxy.model.tags import GalaxyTagHandler
from galaxy.objectstore import BaseObjectStore
from galaxy.quota import QuotaAgent
from galaxy.security.idencoding import IdEncodingHelper
from galaxy.security.vault import Vault
from galaxy.tool_shed.cache import ToolShedRepositoryCache
from galaxy.tool_util.data import ToolDataTableManager
from galaxy.tool_util.deps.containers import ContainerFinder
from galaxy.tool_util.deps.views import DependencyResolversView
from galaxy.tool_util.verify import test_data
from galaxy.util.tool_shed.tool_shed_registry import Registry as ToolShedRegistry
from galaxy.web_stack import ApplicationStack
from galaxy.webhooks import WebhooksRegistry
from galaxy.workflow.trs_proxy import TrsProxy

if TYPE_CHECKING:
    from galaxy.config_watchers import ConfigWatchers
    from galaxy.jobs import JobConfiguration
    from galaxy.managers.collections import DatasetCollectionManager
    from galaxy.managers.hdas import HDAManager
    from galaxy.managers.histories import HistoryManager
    from galaxy.managers.workflows import WorkflowsManager
    from galaxy.tool_shed.galaxy_install.installed_repository_manager import InstalledRepositoryManager
    from galaxy.tools import ToolBox
    from galaxy.tools.cache import ToolCache
    from galaxy.tools.error_reports import ErrorReports
    from galaxy.visualization.genomes import Genomes


class BasicSharedApp(Container):
    """Stripped down version of the ``app`` shared between Galaxy and ToolShed.

    Code that is shared between Galaxy and the Tool Shed should be annotated as
    using BasicSharedApp instead of StructuredApp below.
    """

    name: str
    config: Any  # 'galaxy.config.BaseAppConfiguration'
    datatypes_registry: Registry
    application_stack: ApplicationStack
    model: SharedModelMapping
    security: IdEncodingHelper
    auth_manager: AuthManager
    security_agent: Any
    quota_agent: QuotaAgent

    @property
    def toolbox(self) -> "ToolBox":
        raise NotImplementedError()


class MinimalToolApp(Protocol):
    is_webapp: bool
    name: str
    # Leave config as Any: in a full Galaxy app this is a GalaxyAppConfiguration object, but this is mostly dynamically
    # generated, and here we want to also allow other kinds of configuration objects (e.g. a Bunch).
    config: Any
    datatypes_registry: Registry
    object_store: BaseObjectStore
    tool_data_tables: "ToolDataTableManager"
    file_sources: ConfiguredFileSources
    security: IdEncodingHelper


class MinimalApp(BasicSharedApp):
    is_webapp: bool  # is_webapp will be set to true when building WSGI app
    tag_handler: GalaxyTagHandler
    model: GalaxyModelMapping
    install_model: ModelMapping
    security_agent: GalaxyRBACAgent
    host_security_agent: HostAgent
    server_starttime: int


class MinimalManagerApp(MinimalApp):
    # Minimal App that is sufficient to run Celery tasks
    file_sources: ConfiguredFileSources
    genome_builds: GenomeBuilds
    dataset_collection_manager: "DatasetCollectionManager"
    tool_data_tables: "ToolDataTableManager"
    history_manager: "HistoryManager"
    hda_manager: "HDAManager"
    workflow_manager: "WorkflowsManager"
    workflow_contents_manager: Any  # 'galaxy.managers.workflows.WorkflowContentsManager'
    library_folder_manager: Any  # 'galaxy.managers.folders.FolderManager'
    library_manager: Any  # 'galaxy.managers.libraries.LibraryManager'
    role_manager: Any  # 'galaxy.managers.roles.RoleManager'
    installed_repository_manager: "InstalledRepositoryManager"
    user_manager: Any
    job_config: "JobConfiguration"
    job_manager: Any  # galaxy.jobs.manager.JobManager
    job_metrics: JobMetrics
    dynamic_tool_manager: Any  # 'galaxy.managers.tools.DynamicToolManager'
    genomes: "Genomes"
    error_reports: "ErrorReports"
    notification_manager: Any  # 'galaxy.managers.notification.NotificationManager'
    object_store: BaseObjectStore
    tool_shed_registry: ToolShedRegistry

    @property
    @abc.abstractmethod
    def is_job_handler(self) -> bool: ...

    def wait_for_toolbox_reload(self, old_toolbox: "ToolBox") -> None: ...


class StructuredApp(MinimalManagerApp):
    """Interface defining typed description of the Galaxy UniverseApplication.

    Ideally nothing that depends on StructuredApp should require
    StructuredApp so we can have a clean import dag. This will
    require setting up a lot more distinction between interfaces
    and implementations in Galaxy though. In the meantime, for
    imports that would bring in StructuredApp if properly type
    (cyclical imports), we're just setting the class attributes to
    Any.
    """

    amqp_internal_connection_obj: Optional[Connection]
    dependency_resolvers_view: DependencyResolversView
    container_finder: ContainerFinder
    tool_dependency_dir: Optional[str]
    test_data_resolver: test_data.TestDataResolver
    trs_proxy: TrsProxy
    vault: Vault
    webhooks_registry: WebhooksRegistry
    queue_worker: Any  # 'galaxy.queue_worker.GalaxyQueueWorker'
    data_provider_registry: Any  # 'galaxy.visualization.data_providers.registry.DataProviderRegistry'
    tool_data_tables: ToolDataTableManager
    tool_cache: "ToolCache"
    tool_shed_repository_cache: Optional[ToolShedRepositoryCache]
    watchers: "ConfigWatchers"
    workflow_scheduling_manager: Any  # 'galaxy.workflow.scheduling_manager.WorkflowSchedulingManager'
    interactivetool_manager: Any
    api_keys_manager: Any  # 'galaxy.managers.api_keys.ApiKeyManager'
    visualizations_registry: Any  # 'galaxy.visualization.plugins.registry.VisualizationsRegistry'



"""
Mock infrastructure for testing ModelManagers.
"""

import os
import shutil
import tempfile
from typing import (
    Any,
    cast,
    Optional,
)

from galaxy import (
    di,
    quota,
)
from galaxy.app import UniverseApplication
from galaxy.auth import AuthManager
from galaxy.celery import set_thread_app
from galaxy.config import CommonConfigurationMixin
from galaxy.config_watchers import ConfigWatchers
from galaxy.job_metrics import JobMetrics
from galaxy.jobs.manager import NoopManager
from galaxy.managers.collections import DatasetCollectionManager
from galaxy.managers.dbkeys import GenomeBuilds
from galaxy.managers.hdas import HDAManager
from galaxy.managers.histories import HistoryManager
from galaxy.managers.jobs import JobSearch
from galaxy.managers.users import UserManager
from galaxy.managers.workflows import WorkflowsManager
from galaxy.model import tags
from galaxy.model.base import (
    ModelMapping,
    SharedModelMapping,
    transaction,
)
from galaxy.model.mapping import GalaxyModelMapping
from galaxy.model.scoped_session import galaxy_scoped_session
from galaxy.model.unittest_utils import (
    GalaxyDataTestApp,
    GalaxyDataTestConfig,
)
from galaxy.security import idencoding
from galaxy.short_term_storage import (
    ShortTermStorageAllocator,
    ShortTermStorageConfiguration,
    ShortTermStorageManager,
    ShortTermStorageMonitor,
)
from galaxy.structured_app import (
    BasicSharedApp,
    MinimalManagerApp,
    StructuredApp,
)
from galaxy.tool_util.deps.containers import NullContainerFinder
from galaxy.tools import ToolBox
from galaxy.tools.cache import ToolCache
from galaxy.tools.data import ToolDataTableManager
from galaxy.util import StructuredExecutionTimer
from galaxy.util.bunch import Bunch
from galaxy.web_stack import ApplicationStack


# =============================================================================
def buildMockEnviron(**kwargs):
    environ = {
        "CONTENT_LENGTH": "0",
        "CONTENT_TYPE": "",
        "HTTP_ACCEPT": "*/*",
        "HTTP_ACCEPT_ENCODING": "gzip, deflate",
        "HTTP_ACCEPT_LANGUAGE": "en-US,en;q=0.8,zh;q=0.5,ja;q=0.3",
        "HTTP_CACHE_CONTROL": "no-cache",
        "HTTP_CONNECTION": "keep-alive",
        "HTTP_DNT": "1",
        "HTTP_HOST": "localhost:8000",
        "HTTP_ORIGIN": "http://localhost:8000",
        "HTTP_PRAGMA": "no-cache",
        "HTTP_REFERER": "http://localhost:8000",
        "HTTP_USER_AGENT": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10.10; rv:43.0) Gecko/20100101 Firefox/43.0",
        "PATH_INFO": "/",
        "QUERY_STRING": "",
        "REMOTE_ADDR": "127.0.0.1",
        "REQUEST_METHOD": "GET",
        "SCRIPT_NAME": "",
        "SERVER_NAME": "127.0.0.1",
        "SERVER_PORT": "8080",
        "SERVER_PROTOCOL": "HTTP/1.1",
    }
    environ.update(**kwargs)
    return environ


class MockApp(di.Container, GalaxyDataTestApp):
    config: "MockAppConfig"
    amqp_type: str
    job_search: Optional[JobSearch] = None
    _toolbox: ToolBox
    tool_cache: ToolCache
    install_model: ModelMapping
    watchers: ConfigWatchers
    dataset_collection_manager: DatasetCollectionManager
    hda_manager: HDAManager
    workflow_manager: WorkflowsManager
    history_manager: HistoryManager
    job_metrics: JobMetrics
    stop: bool
    is_webapp: bool = True

    def __init__(self, config=None, **kwargs) -> None:
        super().__init__()
        config = config or MockAppConfig(**kwargs)
        GalaxyDataTestApp.__init__(self, config=config, **kwargs)
        self.install_model = self.model
        self[BasicSharedApp] = cast(BasicSharedApp, self)
        self[MinimalManagerApp] = cast(MinimalManagerApp, self)  # type: ignore[type-abstract]
        self[StructuredApp] = cast(StructuredApp, self)  # type: ignore[type-abstract]
        self[idencoding.IdEncodingHelper] = self.security
        self.name = kwargs.get("name", "galaxy")
        self[SharedModelMapping] = self.model
        self[GalaxyModelMapping] = self.model
        sts_config = ShortTermStorageConfiguration(short_term_storage_directory=os.path.join(config.data_dir, "sts"))
        sts_manager = ShortTermStorageManager(sts_config)
        self[ShortTermStorageAllocator] = sts_manager  # type: ignore[type-abstract]
        self[ShortTermStorageMonitor] = sts_manager  # type: ignore[type-abstract]
        self[galaxy_scoped_session] = self.model.context
        self.visualizations_registry = MockVisualizationsRegistry()
        self.tag_handler = tags.GalaxyTagHandler(self.model.context)
        self[tags.GalaxyTagHandler] = self.tag_handler
        self.quota_agent = quota.DatabaseQuotaAgent(self.model)
        self.job_config = Bunch(
            dynamic_params=None,
            destinations={},
            assign_handler=lambda *args, **kwargs: None,
            get_job_tool_configurations=lambda ids, tool_classes: [Bunch(handler=Bunch())],
        )
        self.tool_data_tables = ToolDataTableManager(tool_data_path=self.config.tool_data_path)
        self.dataset_collections_service = None
        self.container_finder = NullContainerFinder()
        self._toolbox_lock = MockLock()
        self.tool_shed_registry = Bunch(tool_sheds={})
        self.genome_builds = GenomeBuilds(self)
        self.job_manager = NoopManager()
        self.application_stack = ApplicationStack()
        self.auth_manager = AuthManager(self.config)
        self.user_manager = UserManager(cast(BasicSharedApp, self))
        self.execution_timer_factory = Bunch(get_timer=StructuredExecutionTimer)
        self.interactivetool_manager = Bunch(create_interactivetool=lambda *args, **kwargs: None)
        self.is_job_handler = False
        self.biotools_metadata_source = None
        set_thread_app(self)

        def url_for(*args, **kwds):
            return "/mock/url"

        self.url_for = url_for

    @property
    def toolbox(self) -> ToolBox:
        return self._toolbox

    @toolbox.setter
    def toolbox(self, toolbox: ToolBox):
        self._toolbox = toolbox

    def wait_for_toolbox_reload(self, toolbox):
        # TODO: If the tpm test case passes, does the operation really
        # need to wait.
        return True

    def reindex_tool_search(self) -> None:
        raise NotImplementedError


class MockLock:
    def __enter__(self):
        pass

    def __exit__(self, type, value, traceback):
        pass


class MockAppConfig(GalaxyDataTestConfig, CommonConfigurationMixin):
    class MockSchema(Bunch):
        defaults = {"tool_dependency_dir": "dependencies"}

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.schema = self.MockSchema()
        self.use_remote_user = kwargs.get("use_remote_user", False)
        self.enable_celery_tasks = False
        self.tool_data_path = os.path.join(self.root, "tool-data")
        self.galaxy_data_manager_data_path = self.tool_data_path
        self.tool_dependency_dir = None
        self.metadata_strategy = "directory"

        self.user_activation_on = False
        self.new_user_dataset_access_role_default_private = False

        self.activation_grace_period = 0
        self.allow_user_dataset_purge = True
        self.allow_user_creation = True
        self.auth_config_file = "config/auth_conf.xml.sample"
        self.custom_activation_email_message = "custom_activation_email_message"
        self.email_domain_allowlist_content = None
        self.email_domain_blocklist_content = None
        self.email_from = "email_from"
        self.enable_old_display_applications = True
        self.error_email_to = "admin@email.to"
        self.expose_dataset_path = True
        self.hostname = "hostname"
        self.instance_resource_url = "instance_resource_url"
        self.password_expiration_period = 0
        self.pretty_datetime_format = "pretty_datetime_format"
        self.redact_username_in_logs = False
        self.smtp_server = True
        self.terms_url = "terms_url"
        self.templates_dir = "templates"

        self.umask = 0o77
        self.flush_per_n_datasets = 0

        # Compliance related config
        self.redact_email_in_job_name = False

        # Follow two required by GenomeBuilds
        self.len_file_path = os.path.join("tool-data", "shared", "ucsc", "chrom")
        self.builds_file_path = os.path.join("tool-data", "shared", "ucsc", "builds.txt.sample")

        self.shed_tool_config_file = "config/shed_tool_conf.xml"
        self.shed_tool_config_file_set = False
        self.update_integrated_tool_panel = True
        self.enable_beta_edam_toolbox = False
        self.preserve_python_environment = "always"
        self.enable_beta_gdpr = False

        self.version_major = "19.09"

        # set by MockDir
        self.enable_tool_document_cache = False
        self.tool_cache_data_dir = os.path.join(self.root, "tool_cache")
        self.external_chown_script = None
        self.check_job_script_integrity = False
        self.check_job_script_integrity_count = 0
        self.check_job_script_integrity_sleep = 0

        self.default_panel_view = "default"
        self.panel_views_dir = ""
        self.panel_views = {}
        self.edam_panel_views = ""

        self.config_file = None

        self._admin_users = ""
        self.drmaa_external_runjob_script = None
        self.tool_secret = None
        self.track_jobs_in_database = False
        self.amqp_internal_connection = None
        self.tool_configs = []
        self.manage_dependency_relationships = False
        self.enable_tool_shed_check = False
        self.monitor_thread_join_timeout = 1
        self.integrated_tool_panel_config = None
        self.vault_config_file = kwargs.get("vault_config_file")
        self.max_discovered_files = 10000
        self.display_builtin_converters = True
        self.enable_notification_system = True

    @property
    def config_dict(self):
        return self.dict()

    def __getattr__(self, name):
        # Handle the automatic [option]_set options: for tests, assume none are set
        if name == "is_set":
            return lambda x: False
        # Handle the automatic config file _set options
        if name.endswith("_file_set"):
            return False
        raise AttributeError(name)


class MockWebapp:
    def __init__(self, security: idencoding.IdEncodingHelper, **kwargs):
        self.name = kwargs.get("name", "galaxy")
        self.security = security


def mock_url_builder(*a, **k):
    return f"(fake url): {a}, {k}"


class MockTrans:
    def __init__(self, app=None, user=None, history=None, **kwargs):
        self.app = cast(UniverseApplication, app or MockApp(**kwargs))
        self.model = self.app.model
        self.webapp = MockWebapp(self.app.security, **kwargs)
        self.sa_session = self.app.model.session
        self.workflow_building_mode = False
        self.error_message = None
        self.anonymous = False
        self.debug = True
        self.user_is_admin = True
        self.url_builder = mock_url_builder

        self.galaxy_session = None
        self.__user = user
        self.security = self.app.security
        self.history = history

        self.request: Any = Bunch(headers={}, is_body_readable=False, host="request.host")
        self.response: Any = Bunch(headers={}, set_content_type=lambda i: None)

    @property
    def tag_handler(self):
        return self.app.tag_handler

    def check_csrf_token(self, payload):
        pass

    def handle_user_login(self, user):
        pass

    def log_event(self, message):
        pass

    def get_user(self):
        if self.galaxy_session:
            return self.galaxy_session.user
        else:
            return self.__user

    def set_user(self, user):
        """Set the current user."""
        if self.galaxy_session:
            self.galaxy_session.user = user
            self.sa_session.add(self.galaxy_session)
            with transaction(self.sa_session):
                self.sa_session.commit()
        self.__user = user

    user = property(get_user, set_user)

    def get_history(self, **kwargs):
        return self.history

    def set_history(self, history):
        self.history = history

    def fill_template(self, filename, template_lookup=None, **kwargs):
        template = template_lookup.get_template(filename)
        kwargs.update(h=MockTemplateHelpers())
        return template.render(**kwargs)


class MockVisualizationsRegistry:
    BUILT_IN_VISUALIZATIONS = ["trackster"]

    def get_visualizations(self, trans, target):
        return []


class MockDir:
    def __init__(self, structure_dict, where=None):
        self.structure_dict = structure_dict
        self.create_root(structure_dict, where)

    def create_root(self, structure_dict, where=None):
        self.root_path = tempfile.mkdtemp(dir=where)
        self.create_structure(self.root_path, structure_dict)

    def create_structure(self, current_path, structure_dict):
        for k, v in structure_dict.items():
            # if value is string, create a file in the current path and write v as file contents
            if isinstance(v, str):
                self.create_file(os.path.join(current_path, k), v)
            # if it's a dict, create a dir here named k and recurse into it
            if isinstance(v, dict):
                subdir_path = os.path.join(current_path, k)
                os.mkdir(subdir_path)
                self.create_structure(subdir_path, v)

    def create_file(self, path, contents):
        with open(path, "w") as newfile:
            newfile.write(contents)

    def remove(self):
        shutil.rmtree(self.root_path)


class MockTemplateHelpers:
    def js(*js_files):
        pass

    def css(*css_files):
        pass



# We put a tool that references this package into the tool shed
# so we have to provide this legacy location for import indefinitely
# it seems.
from galaxy.util.ucsc import (
    UCSCLimitException,
    UCSCOutWrapper,
)

__all__ = ("UCSCOutWrapper", "UCSCLimitException")



from galaxy.exceptions import (
    NoContentException,
    ObjectNotFound,
)
from galaxy.managers import base
from galaxy.managers.context import ProvidesAppContext
from galaxy.model.base import transaction
from galaxy.schema.fields import DecodedDatabaseIdField
from galaxy.schema.item_tags import ItemTagsCreatePayload
from galaxy.structured_app import MinimalManagerApp


class ItemTagsManager:
    """Interface/service object shared by controllers for interacting with item-tags."""

    def __init__(self, app: MinimalManagerApp) -> None:
        self._app = app
        self._tag_handler = app.tag_handler

    def index(self, trans: ProvidesAppContext, tagged_item_class: str, tagged_item_id: DecodedDatabaseIdField):
        """Displays a collection (list) of tags associated with an item."""
        tags = self._get_user_tags(trans, tagged_item_class, tagged_item_id)
        return [self._api_value(tag, trans, view="collection") for tag in tags]

    def show(
        self, trans: ProvidesAppContext, tagged_item_class: str, tagged_item_id: DecodedDatabaseIdField, tag_name: str
    ):
        """Displays information about a tag associated with an item."""
        tag = self._get_item_tag_assoc(trans, tagged_item_class, tagged_item_id, tag_name)
        if not tag:
            raise ObjectNotFound("Failed to retrieve specified tag.")
        return self._api_value(tag, trans)

    def create(
        self,
        trans: ProvidesAppContext,
        tagged_item_class: str,
        tagged_item_id: DecodedDatabaseIdField,
        tag_name: str,
        payload: ItemTagsCreatePayload,
    ):
        """Apply a new set of tags to an item."""
        value = payload.value
        tag = self._apply_item_tag(trans, tagged_item_class, tagged_item_id, tag_name, value)
        return self._api_value(tag, trans)

    def delete(
        self, trans: ProvidesAppContext, tagged_item_class: str, tagged_item_id: DecodedDatabaseIdField, tag_name: str
    ):
        """Remove a tag from an item."""
        deleted = self._remove_items_tag(trans, tagged_item_class, tagged_item_id, tag_name)
        if not deleted:
            raise NoContentException("Failed to delete specified tag.")
        return deleted

    def _get_tagged_item(self, trans, item_class_name, id, check_ownership=True):
        tagged_item = base.get_object(
            trans, id, item_class_name, check_ownership=check_ownership, check_accessible=True
        )
        return tagged_item

    def _get_user_tags(self, trans, item_class_name, id):
        user = trans.user
        tagged_item = self._get_tagged_item(trans, item_class_name, id)
        return [tag for tag in tagged_item.tags if tag.user == user]

    def _remove_items_tag(self, trans, item_class_name, id, tag_name):
        user = trans.user
        tagged_item = self._get_tagged_item(trans, item_class_name, id)
        deleted = tagged_item and self._tag_handler.remove_item_tag(user, tagged_item, tag_name)
        with transaction(trans.sa_session):
            trans.sa_session.commit()
        return deleted

    def _apply_item_tag(self, trans, item_class_name, id, tag_name, tag_value=None):
        user = trans.user
        tagged_item = self._get_tagged_item(trans, item_class_name, id)
        tag_assoc = self._tag_handler.apply_item_tag(user, tagged_item, tag_name, tag_value)
        with transaction(trans.sa_session):
            trans.sa_session.commit()
        return tag_assoc

    def _get_item_tag_assoc(self, trans, item_class_name, id, tag_name):
        user = trans.user
        tagged_item = self._get_tagged_item(trans, item_class_name, id)
        return self._tag_handler._get_item_tag_assoc(user, tagged_item, tag_name)

    def _api_value(self, tag, trans, view="element"):
        return tag.to_dict(view=view, value_mapper={"id": trans.security.encode_id})



import logging
import os
from typing import (
    Any,
    Dict,
    Optional,
)

from sqlalchemy import false

from galaxy import util
from galaxy.model.base import transaction
from galaxy.model.tool_shed_install import ToolShedRepository
from galaxy.tool_shed.galaxy_install.client import InstallationTarget
from galaxy.tool_shed.galaxy_install.tools import tool_panel_manager
from galaxy.tool_shed.metadata.metadata_generator import GalaxyMetadataGenerator
from galaxy.tool_shed.util.repository_util import (
    get_installed_tool_shed_repository,
    get_repository_owner,
)
from galaxy.tool_shed.util.tool_util import generate_message_for_invalid_tools
from galaxy.util import inflector
from galaxy.util.tool_shed import (
    common_util,
    xml_util,
)
from galaxy.web.form_builder import SelectField

log = logging.getLogger(__name__)


class InstalledRepositoryMetadataManager(GalaxyMetadataGenerator):
    app: InstallationTarget

    def __init__(
        self,
        app: InstallationTarget,
        tpm: Optional[tool_panel_manager.ToolPanelManager] = None,
        repository: Optional[ToolShedRepository] = None,
        changeset_revision: Optional[str] = None,
        repository_clone_url: Optional[str] = None,
        shed_config_dict: Optional[Dict[str, Any]] = None,
        relative_install_dir: Optional[str] = None,
        repository_files_dir: Optional[str] = None,
        resetting_all_metadata_on_repository: bool = False,
        updating_installed_repository: bool = False,
        persist: bool = False,
        metadata_dict: Optional[Dict[str, Any]] = None,
    ):
        super().__init__(
            app,
            repository,
            changeset_revision,
            repository_clone_url,
            shed_config_dict,
            relative_install_dir,
            repository_files_dir,
            resetting_all_metadata_on_repository,
            updating_installed_repository,
            persist,
            metadata_dict=metadata_dict,
            user=None,
        )
        if tpm is None:
            self.tpm = tool_panel_manager.ToolPanelManager(self.app)
        else:
            self.tpm = tpm

    def build_repository_ids_select_field(self, name="repository_ids", multiple=True, display="checkboxes"):
        """Generate the current list of repositories for resetting metadata."""
        repositories_select_field = SelectField(name=name, multiple=multiple, display=display)
        query = self.get_query_for_setting_metadata_on_repositories(order=True)
        for repository in query:
            owner = str(repository.owner)
            option_label = f"{str(repository.name)} ({owner})"
            option_value = f"{self.app.security.encode_id(repository.id)}"
            repositories_select_field.add_option(option_label, option_value)
        return repositories_select_field

    def get_query_for_setting_metadata_on_repositories(self, order=True):
        """
        Return a query containing repositories for resetting metadata.  The order parameter
        is used for displaying the list of repositories ordered alphabetically for display on
        a page.  When called from the Galaxy API, order is False.
        """
        if order:
            return (
                self.app.install_model.context.query(self.app.install_model.ToolShedRepository)
                .filter(self.app.install_model.ToolShedRepository.table.c.uninstalled == false())
                .order_by(
                    self.app.install_model.ToolShedRepository.table.c.name,
                    self.app.install_model.ToolShedRepository.table.c.owner,
                )
            )
        else:
            return self.app.install_model.context.query(self.app.install_model.ToolShedRepository).filter(
                self.app.install_model.ToolShedRepository.table.c.uninstalled == false()
            )

    def get_repository_tools_tups(self):
        """
        Return a list of tuples of the form (relative_path, guid, tool) for each tool defined
        in the received tool shed repository metadata.
        """
        repository_tools_tups = []
        shed_conf_dict = self.tpm.get_shed_tool_conf_dict(self.metadata_dict.get("shed_config_filename"))
        if "tools" in self.metadata_dict:
            for tool_dict in self.metadata_dict["tools"]:
                load_relative_path = relative_path = tool_dict.get("tool_config", None)
                if shed_conf_dict.get("tool_path"):
                    load_relative_path = os.path.join(shed_conf_dict.get("tool_path"), relative_path)
                guid = tool_dict.get("guid", None)
                if relative_path and guid:
                    try:
                        tool = self.app.toolbox.load_tool(
                            os.path.abspath(load_relative_path), guid=guid, use_cached=False
                        )
                    except Exception:
                        log.exception("Error while loading tool at path '%s'", load_relative_path)
                        tool = None
                else:
                    tool = None
                if tool:
                    repository_tools_tups.append((relative_path, guid, tool))
        return repository_tools_tups

    def reset_all_metadata_on_installed_repository(self):
        """Reset all metadata on a single tool shed repository installed into a Galaxy instance."""
        if self.relative_install_dir:
            assert self.repository
            original_metadata_dict = self.repository.metadata_
            self.generate_metadata_for_changeset_revision()
            if self.metadata_dict != original_metadata_dict:
                self.repository.metadata_ = self.metadata_dict
                self.update_in_shed_tool_config()

                session = self.app.install_model.context
                session.add(self.repository)
                with transaction(session):
                    session.commit()

                log.debug(f"Metadata has been reset on repository {self.repository.name}.")
            else:
                log.debug(f"Metadata did not need to be reset on repository {self.repository.name}.")
        else:
            log.debug(
                f"Error locating installation directory for repository {self.repository and self.repository.name}."
            )

    def reset_metadata_on_selected_repositories(self, user, **kwd):
        """
        Inspect the repository changelog to reset metadata for all appropriate changeset revisions.
        This method is called from both Galaxy and the Tool Shed.
        """
        message = ""
        status = "done"
        if repository_ids := util.listify(kwd.get("repository_ids", None)):
            successful_count = 0
            unsuccessful_count = 0
            for repository_id in repository_ids:
                try:
                    repository = get_installed_tool_shed_repository(self.app, repository_id)
                    self.set_repository(repository)
                    self.reset_all_metadata_on_installed_repository()
                    if self.invalid_file_tups:
                        message = generate_message_for_invalid_tools(
                            self.app, self.invalid_file_tups, repository, None, as_html=False
                        )
                        log.debug(message)
                        unsuccessful_count += 1
                    else:
                        log.debug(
                            "Successfully reset metadata on repository %s owned by %s",
                            repository.name,
                            repository.owner,
                        )
                        successful_count += 1
                except Exception:
                    log.exception("Error attempting to reset metadata on repository %s", str(repository.name))
                    unsuccessful_count += 1
            message = "Successfully reset metadata on %d %s.  " % (
                successful_count,
                inflector.cond_plural(successful_count, "repository"),
            )
            if unsuccessful_count:
                message += "Error setting metadata on %d %s - see the galaxy log for details.  " % (
                    unsuccessful_count,
                    inflector.cond_plural(unsuccessful_count, "repository"),
                )
        else:
            message = "Select at least one repository to on which to reset all metadata."
            status = "error"
        return message, status

    def set_repository(self, repository):
        super().set_repository(repository)
        self.repository_clone_url = common_util.generate_clone_url_for_installed_repository(self.app, repository)

    def tool_shed_from_repository_clone_url(self) -> str:
        """Given a repository clone URL, return the tool shed that contains the repository."""
        repository_clone_url = self.repository_clone_url
        assert repository_clone_url
        cleaned_repository_clone_url = common_util.remove_protocol_and_user_from_clone_url(repository_clone_url)
        return (
            common_util.remove_protocol_and_user_from_clone_url(cleaned_repository_clone_url)
            .split("/repos/")[0]
            .rstrip("/")
        )

    def update_in_shed_tool_config(self):
        """
        A tool shed repository is being updated so change the shed_tool_conf file.  Parse the config
        file to generate the entire list of config_elems instead of using the in-memory list.
        """
        assert self.repository
        shed_conf_dict = self.shed_config_dict or self.repository.get_shed_config_dict(self.app)
        shed_tool_conf = shed_conf_dict["config_filename"]
        tool_path = shed_conf_dict["tool_path"]
        self.tpm.generate_tool_panel_dict_from_shed_tool_conf_entries(self.repository)
        repository_tools_tups = self.get_repository_tools_tups()
        clone_url = common_util.generate_clone_url_for_installed_repository(self.app, self.repository)
        tool_shed = self.tool_shed_from_repository_clone_url()
        owner = self.repository.owner
        if not owner:
            cleaned_repository_clone_url = common_util.remove_protocol_and_user_from_clone_url(clone_url)
            owner = get_repository_owner(cleaned_repository_clone_url)
        guid_to_tool_elem_dict = {}
        for tool_config_filename, guid, tool in repository_tools_tups:
            guid_to_tool_elem_dict[guid] = self.tpm.generate_tool_elem(
                tool_shed,
                self.repository.name,
                self.repository.changeset_revision,
                self.repository.owner or "",
                tool_config_filename,
                tool,
                None,
            )
        config_elems = []
        tree, error_message = xml_util.parse_xml(shed_tool_conf)
        if tree:
            root = tree.getroot()
            for elem in root:
                if elem.tag == "section":
                    for i, tool_elem in enumerate(elem):
                        guid = tool_elem.attrib.get("guid")
                        if guid in guid_to_tool_elem_dict:
                            elem[i] = guid_to_tool_elem_dict[guid]
                elif elem.tag == "tool":
                    guid = elem.attrib.get("guid")
                    if guid in guid_to_tool_elem_dict:
                        elem = guid_to_tool_elem_dict[guid]
                config_elems.append(elem)
            self.tpm.config_elems_to_xml_file(config_elems, shed_tool_conf, tool_path)



"""This module defines the error reporting framework for Galaxy jobs.
"""

import collections
import logging
import os

from galaxy.util import plugin_config

log = logging.getLogger(__name__)

DEFAULT_CONFIG = [
    {
        "type": "email",
        "verbose": True,
        "user_submission": True,
    },
]
DEFAULT_PLUGINS_SOURCE = plugin_config.PluginConfigSource("dict", DEFAULT_CONFIG)


class ErrorReports:
    """Load and store a collection of :class:`ErrorPlugin` objects."""

    def __init__(self, conf_file=None, **kwargs):
        """Load :class:`ErrorPlugin` objects from specified configuration file."""
        self.plugin_classes = self.__plugins_dict()
        self.default_error_plugin = ErrorPlugin.from_file(self.plugin_classes, conf_file, **kwargs)
        self.error_plugin = collections.defaultdict(lambda: self.default_error_plugin)

    def __plugins_dict(self):
        import galaxy.tools.error_reports.plugins

        return plugin_config.plugins_dict(galaxy.tools.error_reports.plugins, "plugin_type")


class NullErrorPlugin:
    def submit_report(self, dataset, job, tool, **kwargs):
        log.warning("Bug report for dataset %s, job %s submitted to NullErrorPlugin", dataset, job)
        return [("Error reporting is not configured for this Galaxy instance", "danger")]


NULL_ERROR_PLUGIN = NullErrorPlugin()


class ErrorPlugin:
    def __init__(self, plugin_classes, plugins_source, **kwargs):
        self.extra_kwargs = kwargs
        self.app = kwargs["app"]
        self.plugin_classes = plugin_classes
        self.plugins = self.__plugins_from_source(plugins_source)

    def _can_access_dataset(self, dataset, user):
        if user:
            roles = user.all_roles()
        else:
            roles = []
        return self.app.security_agent.can_access_dataset(roles, dataset.dataset)

    def submit_report(self, dataset, job, tool, user=None, user_submission=False, **kwargs):
        if user_submission:
            assert self._can_access_dataset(dataset, user), Exception("You are not allowed to access this dataset.")

        responses = []
        for plugin in self.plugins:
            if user_submission == plugin.user_submission:
                try:
                    response = plugin.submit_report(dataset, job, tool, **kwargs)
                    log.debug("Bug report plugin %s generated response %s", plugin, response)
                    if plugin.verbose and response:
                        responses.append(response)
                except Exception:
                    log.exception("Failed to generate submit_report commands for plugin %s", plugin)
        return responses

    def __plugins_from_source(self, plugins_source):
        return plugin_config.load_plugins(self.plugin_classes, plugins_source, self.extra_kwargs)

    @staticmethod
    def from_file(plugin_classes, conf_file, **kwargs):
        plugins_source = DEFAULT_PLUGINS_SOURCE
        if conf_file and os.path.exists(conf_file):
            plugins_source = plugin_config.plugin_source_from_path(conf_file)
        if not plugins_source.source:
            return NULL_ERROR_PLUGIN
        return ErrorPlugin(plugin_classes, plugins_source, **kwargs)
